﻿using System;
using System.IO;

using TeleVox_TeleToken;
using ExCommon;

namespace ExRand
{
	class Class
	{
		[STAThread]
		static void Main(string[] args)
		{
			const string RandInitFileName = "MyRandInitFile.bin";
			byte[][] RandInitTbl = new byte[4][] {
				new byte[16],
				new byte[16],
				new byte[16],
				new byte[16]
			};
			uint n, Time, TTime;
			byte[] RandTbl = new byte[16];

			// przygotowanie tablicy inicjującej wypełnionej możliwie nieprzewidywalną wartościami
			//  - metoda przykładowa prosta ale nie zalecana, należy zastosować własną metodę
			n = 0;
			while (n<TeleToken.TVTT_RAND_INIT_TBL_SIZE) {
				RandInitTbl[n/16][n%16] = (byte) MainDef.GetTickCount();
				n++;
			}

			// odczytanie pliku inicjującego (o ile istnieje)
			try {
				FileStream fsin = new FileStream(RandInitFileName,
					FileMode.Open);
				BinaryReader inb = new BinaryReader(fsin);
				for (n=0;n<4;n++)
					RandInitTbl[n] =
						inb.ReadBytes(TeleToken.TVTT_RAND_INIT_TBL_SIZE / 4);
				// UWAGA! Nie następuje potwierdzenie poprawności odczytu ponieważ jeśli odczyt się nie uda to tablica
				//  jest już wypełniona wartościami możliwie nieprzewidywalnymi
				//  - jest to metoda prosta ale nie zalecana, w rozwiązaniu zalecanym powinno być zaimplementowane:
				//      * sprawdzenie poprawności odczytu
				//      * szyfrowanie pliku z tablicą inicjalizującą
				//      * sprawdzenie oryginalności pliku (suma kontrolna, podpis cyfrowy, etc.)
				inb.Close();
				fsin.Close();
			} catch {};

			// zainicjowanie Generatora liczb pseudolosowych
			TeleToken.TVTT_RandInit(ref RandInitTbl);

			// zapisanie nowej tablicy inicjującej generator
			TeleToken.TVTT_RandGetInit(ref RandInitTbl);
			try {
				FileStream fsout = new FileStream(RandInitFileName,
					FileMode.Create);
				BinaryWriter outb = new BinaryWriter(fsout);
				for (n=0;n<4;n++)
					outb.Write(RandInitTbl[n], 0,
						TeleToken.TVTT_RAND_INIT_TBL_SIZE / 4);
				// UWAGA! nie następuje przeprowadzenie poprawności zapisu
				//  - należy zastosować własną obsługę błędu zapisu zależną od potrzeb
				outb.Close();
				fsout.Close();
			} catch {};

			// powyzsze kroki konieczne żeby program przy następnym uruchomieniu
			// wystartował z innym stanem generatora liczb losowych
			//  - wykonywane na wypadek zakończenia programu bez zapisu ostatniego stanu generatora

			// zbieranie wartości losowych
			// ta operacja powinna być wykonywana w głównej pętli obsługi wiadomości
			// zdarzeniami moga być np.:
			//  * moment wciśnięcia klawisza na klawiaturze (np. odczytany przez GetTickCount())
			//  * kod wciśniętego klawisza
			//  * moment wciśnięcia przycisku myszki (np. odczytany przez GetTickCount())
			//  * moment poruszenia myszki (np. odczytany przez GetTickCount())
			//  * kierunek przesunięcia myszki
			//  * inne mniej lub bardziej nieprzewidywalne zdażenia wystepujące w systemie
			// poniższa procedura może być przydatna przy pierwszym uruchomieniu programu gdy
			// nie ma jeszcze tablicy inicjującej generator liczb pseudolosowych

			Console.Write("Generowanie liczb pseudolosowych: <Enter>\n");
			Console.In.Peek();
			Time = MainDef.GetTickCount() + 1000;
			TTime = MainDef.GetTickCount() + 10000;
			n = 0;
			int ch;
			while (MainDef.GetTickCount()<TTime) {
				n++;
				if (Console.In.Peek() != -1) {                                      // oczekiwanie na wciśnięcie klawisza
					ch = Console.Read();
					TeleToken.TVTT_Randomize(0, MainDef.GetTickCount());        // jako pierwsze źródło przyjęto moment wciśnięcia klawisza
					TeleToken.TVTT_Randomize(1, (uint) ch);                     // jako drugie źródło przyjęto kod wciśniętego klawisza
					TeleToken.TVTT_Randomize(2, (uint) n);                      // jako trzecie źródło przyjęto moment wcisnięcia klawisza wyrażony w ilości obrotów petli
				} else MainDef.Sleep(100);                                          // zwrócenie części czasu do systemu

				if (MainDef.GetTickCount()>Time) {                                  // "co jakiś czas" wygenerowanie i wyświetlenie wartosci pseudolosowej
					Time = MainDef.GetTickCount() + 1000;
					TeleToken.TVTT_Rand(ref RandTbl);
					Console.Write("\t");
					ZeroMem.Print16(RandTbl);
				}
			}

			Console.Write("\n");

			// zapisanie stanu generatora liczb pseudolosowych
			TeleToken.TVTT_RandGetInit(ref RandInitTbl);
			try {
				FileStream fsout = new FileStream(RandInitFileName,
					FileMode.Create);
				BinaryWriter outb = new BinaryWriter(fsout);
				for (n=0;n<4;n++)
					outb.Write(RandInitTbl[n], 0,
						TeleToken.TVTT_RAND_INIT_TBL_SIZE / 4);
				// UWAGA! nie następuje przeprowadzenie poprawności zapisu
				//  - należy zastosować własną obsługę błędu zapisu zależną od potrzeb
				outb.Close();
				fsout.Close();
			} catch {};

			return;
		}
	}
}
