using System;
using System.Runtime.InteropServices;

namespace TeleVox_TeleToken
{
	public class LibWrap {
		[DllImport("Kernel32.dll")]
		static public extern uint GetLastError();

		[DllImport("Kernel32.dll", SetLastError=true)]
		static public extern void SetLastError(uint dwErrCode);

		[DllImport("AdvApi32.dll")]
		static public extern uint RegCreateKeyEx(uint hKey,
												 string lpSubKey,
												 uint Reserved,
												 string lpClass,
												 uint dwOptions,
												 uint samDesired,
												 uint lpSecurityAttributes,
												 ref uint phkResult,
												 ref uint lpdwDisposition);

		[DllImport("AdvApi32.dll")]
		static public extern uint RegQueryValueEx(uint hKey,
												  string lpValueName,
												  uint lpReserved,
												  ref uint lpType,
												  ref byte lpData,
												  ref uint lpcbData);

		[DllImport("AdvApi32.dll")]
		static public extern uint RegSetValueEx(uint hKey,
												string lpValueName,
												uint Reserved,
												uint dwType,
												byte[] lpData,
												uint cbData);

		[DllImport("AdvApi32.dll")]
		static public extern uint RegCloseKey(uint hKey);

		[DllImport("TeleToken.dll")]
		static public extern uint TVTT_GetVersion(ref uint Ver,
												  ref uint SubVer,
												  ref uint ApiVer);

		[DllImport("TeleToken.dll")]
		static public extern uint TVTT_Talk(uint TokenH,
											uint TimeOut,
											uint OutKey,
											ref byte OutDataA,
											ref byte OutDataB,
											ref uint Key,
											ref byte InDataA,
											ref byte InDataB);

		[DllImport("TeleToken.dll")]
		static public extern uint TVTT_FindToken(ref uint TokenH,
												 uint TimeOut,
												 uint OwnerID,
												 byte OwnerMsk,
												 byte Skip,
												 ref uint FeatureTblSize,
												 ref uint FeatureTbl);

		[DllImport("TeleToken.dll")]
		static public extern uint TVTT_StartSession(uint TokenH, uint TimeOut);

		[DllImport("TeleToken.dll")]
		static public extern uint TVTT_EndSession(uint TokenH);

		[DllImport("TeleToken.dll")]
		static public extern uint TVTT_FreeToken(uint TokenH);
	}

	public class TeleToken
	{
		//----------------------------------------------------------------------
		// Stae biblioteki
		public const byte TVTT_BLOCK_A                    = 0;                  /* staa definujca numer bloku danych przy szyfrowaniu pierwszego bloku danych - na potrzeby funkcji generujcej hasa tymczasowe */
		public const byte TVTT_BLOCK_B                    = 1;                  /* staa definujca numer bloku danych przy szyfrowaniu drugiego bloku danych - na potrzeby funkcji generujcej hasa tymczasowe */
		public const byte TVTT_PC2T                       = 0;                  /* staa definujca kierunek przesania wiadomoci z PC do Tokenu - na potrzeby funkcji generujcej hasa tymczasowe */
		public const byte TVTT_T2PC                       = 1;                  /* staa definujca kierunek przesania wiadomoci z Tokenu do PC - na potrzeby funkcji generujcej hasa tymczasowe */

		public const byte TVTT_RAND_INIT_TBL_SIZE         = 16*4;               /* staa definiujca rozmiar tablicy inicjujcej generator liczb pseudolosowych */

		public const byte TVTTS_OK                        = 0;

		//----------------------------------------------------------------------
		// Kody bdw zwracane przez DLL
		public const ushort TVTTS_INVALID_HANDLE          = 0x0100;
		public const ushort TVTTS_NULL_PTR                = 0x0200;
		public const ushort TVTTS_BAD_PARAM               = 0x0300;
		public const ushort TVTTS_TOO_MANY_OPEN_TOKENS    = 0x0400;
		public const ushort TVTTS_USB_DOES_NOT_EXIST      = 0x0500;
		public const ushort TVTTS_TOKEN_NOT_FOUND         = 0x0600;
		public const ushort TVTTS_TIME_OUT                = 0x0700;
		public const ushort TVTTS_SEQ_EXPIRED             = 0x0800;
		public const ushort TVTTS_ALREADY_OPENED          = 0x0900;
		public const ushort TVTTS_UNKNOWN_ERROR           = 0x1000;
		public const ushort TVTTS_DLL_INIT_FAILED         = 0xFF00;

		//----------------------------------------------------------------------
		// kody statusu Tokenu:
		public const byte TVTTS_NO_OPEN                   = 0x11;
		public const byte TVTTS_EXPIRED                   = 0x12;
		public const byte TVTTS_DAMAGED                   = 0x13;

		public const byte TVTTS_NOT_SUPPORTED             = 0x21;
		public const byte TVTTS_OUT_OF_RANGE              = 0x22;
		public const byte TVTTS_PROTECTION_NOT_SUPPORTED  = 0x23;

		public const byte TVTTS_PASS_NOT_PERMITTED        = 0x31;
		public const byte TVTTS_NOT_PERMITTED             = 0x32;
		public const byte TVTTS_ACCESS_EXPIRED            = 0x33;

		//----------------------------------------------------------------------
		// definicja pl tablicy opisujcej moliwoci tokenu
		public const byte TVTTTI_ID                       = 0;
		public const byte TVTTTI_NV_MEM_SIZE              = 1;
		public const byte TVTTTI_RAM_SIZE                 = 2;
		public const byte TVTTTI_PROTECT_MEM_SIZE         = 3;
		public const byte TVTTTI_PASS_TABLE_SIZE          = 4;
		public const byte TVTTTI_BITS_PER_PASS            = 5;
		public const byte TVTTTI_COUNTERS                 = 6;
		public const byte TVTTTI_BITS_PER_COUNTER         = 7;

		public const byte TVTTTI_TABLE_SIZE               = 8;

		//----------------------------------------------------------------------
		// definicja standardowych pl w drugim bloku danych
		public const byte TVTTD_CMD                       = 0;
		public const byte TVTTD_STATUS                    = 0;
		public const byte TVTTD_PARAM1                    = 1;
		public const byte TVTTD_PARAM2                    = 2;
		public const byte TVTTD_PARAM3                    = 3;
		public const byte TVTTD_AUTH                      = 4;

		//----------------------------------------------------------------------
		// kody rozkazw tokenu
		public const byte TVTTC_OPEN                      = 0x01;
		public const byte TVTTC_CLOSE                     = 0x02;
		public const byte TVTTC_FORMAT                    = 0x0F;

		public const byte TVTTC_NVMEM_READ                = 0x11;
		public const byte TVTTC_NVMEM_WRITE               = 0x12;
		public const byte TVTTC_NVMEM_WRITE_PROTECT       = 0x13;
		public const byte TVTTC_NVMEM_PROTECT             = 0x14;
		public const byte TVTTC_VMEM_READ                 = 0x91;
		public const byte TVTTC_VMEM_WRITE                = 0x92;

		public const byte TVTTC_ENCRYPT                   = 0x21;
		public const byte TVTTC_DECRYPT                   = 0x22;

		public const byte TVTTC_CNT_READ                  = 0x31;
		public const byte TVTTC_CNT_INC_READ              = 0x32;
		public const byte TVTTC_CNT_DEC_READ              = 0x33;
		public const byte TVTTC_CNT_WRITE                 = 0x34;
		public const byte TVTTC_CNT_INC_ENCRYPT_READ      = 0x35;
		public const byte TVTTC_CNT_WRITE_CONFIG          = 0x36;
		public const byte TVTTC_CNT_SET_CONFIG            = 0x37;
		public const byte TVTTC_CNT_CLR_CONFIG            = 0x38;

		public const byte TVTTC_PASS_WRITE                = 0x41;
		public const byte TVTTC_PASS_WRITE_CONFIG         = 0x42;
		public const byte TVTTC_PASS_SET_CONFIG           = 0x43;
		public const byte TVTTC_PASS_CLR_CONFIG           = 0x44;

		//----------------------------------------------------------------------
		// bity konfiguracji hase
		public const byte TVTTPC0_TOKEN                   = 0x01;
		public const byte TVTTPC0_PASS_WRITE              = 0x02;
		public const byte TVTTPC0_PASS_CONFIG             = 0x04;
		public const byte TVTTPC0_MEM_READ                = 0x08;
		public const byte TVTTPC0_MEM_WRITE               = 0x10;
		public const byte TVTTPC0_ENCRYPT                 = 0x20;
		public const byte TVTTPC0_DECRYPT                 = 0x40;

		public const byte TVTTPC1_CNT_READ                = 0x01;
		public const byte TVTTPC1_CNT_WRITE               = 0x02;
		public const byte TVTTPC1_CNT_CONFIG              = 0x04;
		public const byte TVTTPC1_ENCRYPT_PASS            = 0x08;
		public const byte TVTTPC1_DECRYPT_PASS            = 0x10;
		public const byte TVTTPC1_CNT_ENCRYPT_PASS        = 0x20;
		public const byte TVTTPC1_WRITE_ENABLED           = 0x40;
		public const byte TVTTPC1_CONFIG_ENABLED          = 0x80;

		//----------------------------------------------------------------------
		// bity konfiguracji licznikw
		public const byte TVTTCC0_WRITE                   = 0x01;
		public const byte TVTTCC0_READ                    = 0x02;
		public const byte TVTTCC0_INC_READ                = 0x04;
		public const byte TVTTCC0_DEC_READ                = 0x08;
		public const byte TVTTCC0_INC_ENCRYPT_READ        = 0x10;
		public const byte TVTTCC0_STOP_AT_ZERO            = 0x20;
		public const byte TVTTCC0_NVMEM_READ              = 0x40;
		public const byte TVTTCC0_NVMEM_WRITE             = 0x80;

		public const byte TVTTCC1_PASS                    = 0x01;
		public const byte TVTTCC1_CONFIG_ENABLED          = 0x80;

		static public byte[] RandKeyA                     = new byte[16];       // generator liczb pseudolosowych - pierwsza poowa klucza generatora
		static public byte[] RandKeyB                     = new byte[16];       // generator liczb pseudolosowych - pierwsza poowa klucza generatora
		static public byte[] RandCnt                      = new byte[16];       // generator liczb pseudolosowych - licznik wywoa generatora liczb pseudolosowych
		static public byte[] RandRand                     = new byte[16*32];    // generator liczb pseudolosowych - bufor na kolejk zdarze losowych
		static public byte[] RandRandNr                   = new byte[256];      // generator liczb pseudolosowych - licznik kolejkowania zdarze losowych
		static public uint[] RandRandCnt                  = new uint[32];       // generator liczb pseudolosowych - licznik zdarze w kolejkach

		public TeleToken()
		{
		}

		/*****************************************************************************
		Funkcja liczy blok kontrolny wiadomoci i umieszcza go w drugim bloku danych
		We:
			Adresy na oba bloki danych
		Wy:
			wyliczony blok autoryzacji zapisany do drugiego bloku danych

		*****************************************************************************/
		static public void TVTT_Calc(byte [] DataA, ref byte [] DataB)
		{
			// wypenienie bloku kontrolnego wartoci 0xFF
			int DstPtr = 0;
			while (DstPtr<12) {
				DataB[TVTTD_AUTH+DstPtr] = 0xFF;
				DstPtr++;
			}

			// doczenie do obszaru kontrolnego, danych pierwszego bloku danych
			DstPtr = 0;
			int SrcPtr=0;
			while (SrcPtr<16) {
				DataB[TVTTD_AUTH+DstPtr] ^= DataA[SrcPtr];
				DstPtr = (DstPtr+1)%12;
				SrcPtr++;
			}

			// doczenie do obszaru kontrolnego, danych drugiego bloku danych
			SrcPtr = 0;
			while (SrcPtr<TVTTD_AUTH) {
				DataB[TVTTD_AUTH+DstPtr] ^= DataB[SrcPtr];
				DstPtr = (DstPtr+1)%12;
				SrcPtr++;
			}
		}

		/*****************************************************************************
		Funkcja sprawdza poprawno bloku kontrolnego wiadomoci
		We:
			Adresy na oba bloki danych
		Wy:
			0 - blok kontrolny poprawny lub !0 - blok kontrolny nie poprawny

		*****************************************************************************/
		static public byte TVTT_Check(byte[] DataA, ref byte[] DataB)
		{
			// "usunicie" z obszaru kontrolnego, danych pierwszego bloku danych
			int DstPtr = 0;
			int SrcPtr = 0;
			while (SrcPtr<16) {
				DataB[TVTTD_AUTH+DstPtr] ^= DataA[SrcPtr];
				DstPtr = (DstPtr+1)%12;
				SrcPtr++;
			}

			// "usunicie" z obszaru kontrolnego, danych drugiego bloku danych
			SrcPtr = 0;
			while (SrcPtr<TVTTD_AUTH) {
				DataB[TVTTD_AUTH+DstPtr] ^= DataB[SrcPtr];
				DstPtr = (DstPtr+1)%12;
				SrcPtr++;
			}

			// sprawdzenie czy blok kontrolny wypeniony wartoci 0xFF
			DstPtr = 0;
			while (DstPtr<12) {
				if (DataB[TVTTD_AUTH+DstPtr] != 0xFF) return(1);                // bd w bloku kontrolnym
				DstPtr++;
			}

			return(0);                                                          // blok kontrolny poprawny
		}

		/*****************************************************************************
		Funkcja wylicza klucz sesji
		We:
			Adres na wyliczony klucz sesji (adres na tablic docelow)
			Has-o PC
			Tablica losowa PC
			Has-o Tokenu
			Tablica losowa Tokenu
		Wy:
			wyliczony klucz sesji wpisany do podanej tablicy

		*****************************************************************************/
		static public void TVTT_CalcSessionKey(ref byte[] SessionKey,
											   byte[] PassPC,
											   byte[] RandPC,
											   byte[] PassT,
											   byte[] RandT)
		{
			byte[] Data = new byte[16];

			int i = 0;                                                          // skopiowanie tablicy losowej PC do bufora
			while(i<16) {
				Data[i] = RandPC[i];
				i++;
			}

			rijndael.rijndaelEncrypt(PassT, ref Data);                          // zaszyfrowanie tablicy losowej PC hasem Tokenu

			i = 0;                                                              // przepisanie uzyskanej czci klucza do tablicy klucza sesji
			while (i<16) {
				SessionKey[i] = Data[i];
				i++;
			}

			i = 0;                                                              // skopiowanie tablicy losowej Tokenu do bufora
			while (i<16) {
				Data[i] = RandT[i];
				i++;
			}

			rijndael.rijndaelEncrypt(PassPC, ref Data);                         // zaszyfrowanie tablicy losowej Tokenu hasem PC

			i = 0;                                                              // poczenie pierwszej czci klucza sesji z uzyskan drug czci klucza sesji
			while (i<16) {
				SessionKey[i] ^= Data[i];
				i++;
			}

			i = 0;                                                              // wyczyszczenie bufora pomocniczego
			while (i<16) {
				Data[i] = 0;
				i++;
			}
		}

		/*****************************************************************************
		Funkcja rozszyfrowuje lub szyfruje podany blok danych
		We:
			Adres na blok do zaszyfrowania
			Klucza szyfrowania wiadomoci
			Has-o sesji
			Numer bloku danych (0-blok pierwszy, 1-blok drugi)
			Kierunek przesy-ania wiadomoci (0 - PC->Token, 1 - Token->PC)
		Wy:
			zaszyfrowany blok danych

		*****************************************************************************/
		static public void TVTT_CryptBlock(ref byte [] Block, uint Key,
										   byte [] Pass, byte BlockNr,
										   byte Direction)
		{
			byte[] HashTable = new byte[16];

			if (Key == 0) return;                                               // Key=0 jest znacznikiem wiadomoci nieszyfrowanej -> wyjcie

			int i = 0;                                                          // wyczyszczenie tablicy mieszajcej
			while (i<16) {
				HashTable[i]=0;
				i++;
			}

			HashTable[0] = (byte) Key;                                          // przygotowanie do generowania klucza tymczasowego - wpisanie klucza szyfrowania wiadomoci do tablicy wynikowej
			Key = Key>>8;
			HashTable[1] = (byte) Key;
			Key = Key>>8;
			HashTable[2] = (byte) Key;

			HashTable[3] = (byte) BlockNr;                                      // przygotowanie do gewnerowania klucza tymczasowego - do-czenie numeru bloku danych
			HashTable[4] = (byte) Direction;                                    // przygotowanie do gewnerowania klucza tymczasowego - do-czenie znacznika kierunku przesy-u wiadomoci
			rijndael.rijndaelEncrypt(Pass, ref HashTable);                      // wygenerowanie hasa tymczasowego

			i=0;                                                                // rozszyfrowanie lub zaszyfrowanie bloku oraz wyczyszczenie tablicy mieszajcej
			while (i<16) {
				Block[i] ^= HashTable[i];
				HashTable[i] = 0;
				i++;
			}
		}

		/*****************************************************************************
		Funkcja inicjuje generator liczb pseudolosowych
		We:
			Adres na blok danych inicjujcych

		*****************************************************************************/
		static public void TVTT_RandInit(ref byte[][] Init)
		{
			uint n;

			// rozszyfrowanie tablic
			n = 1;
			while (n<4) {
				rijndael.rijndaelDecrypt(Init[0], ref Init[n]);
				n++;
			}

			// usunicie mieszania tablic
			n = 16;
			while(n<(16*4)) {
				Init[n/16][n%16] ^= (byte) n;
				n++;
			}

			// przepisanie danych z bloku do tablic
			n=0;
			while (n<16) {
				RandKeyA[n] = Init[1][n];
				RandKeyB[n] = Init[2][n];
				RandCnt[n] = Init[3][n];
				n++;
			}
		}

		/*****************************************************************************
		Funkcja zwraca blok danych do uycia przy nastpnej inicjalizacji generatora
			liczb pseudolosowych
		We:
			Adres na blok do ktrego zapis danych opisujcych stan generatora
		Wy:
			Blok danych opisujcych stan generatora

		*****************************************************************************/
		static public void TVTT_RandGetInit(ref byte[][] Init)
		{
			TVTT_Rand(ref Init[0]);
			TVTT_Rand(ref Init[1]);
			TVTT_Rand(ref Init[2]);
			TVTT_Rand(ref Init[3]);
		}

		/*****************************************************************************
		Funkcja dodaje zdarzenia losowe do generatora liczb pseudolosowych
		We:
			Numer rda zdarze losowych
			Warto zdarzenia

		*****************************************************************************/
		static public void TVTT_Randomize(byte SrcNr, uint Data)
		{
			byte[] Buff = new byte[16];
			byte[] Buf2 = new byte[16];
			byte n;
			uint RandOffset;

			// zwikszenie licznika danych w kolejce
			if (RandRandCnt[RandRandNr[SrcNr&0xFF]&31]!=0xFFFFFFFF)
				(RandRandCnt[RandRandNr[SrcNr&0xFF]&31])++;

			// wyliczenie offsetu na kolejk i zwikszenie licznika kolejek
			RandOffset = (uint)((RandRandNr[SrcNr&0xFF]&31)*16);
			RandRandNr[SrcNr] = (byte)((RandRandNr[SrcNr]+1)&31);

			// wyczyszczenie bufora i wpisanie do niego danych o zdarzeniu
			n = 0;
			while (n<16) {
				Buff[n] = 0;
				n++;
			}

			Buff[0] = SrcNr;
			Buff[1] = (byte) Data;
			Data = Data>>8;
			Buff[2] = (byte) Data;
			Data = Data>>8;
			Buff[3] = (byte) Data;
			Data = Data>>8;
			Buff[4] = (byte) Data;

			// pomieszanie danych o zdarzeniu z danymi z kolejki
			for (n=0; n<16; n++) Buf2[n] = RandRand[n+RandOffset];
			rijndael.rijndaelEncrypt(Buf2, ref Buff);

			// wpisanie nowych danych do kolejki i zamazanie zawartoci bufora
			n = 0;
			while (n<16) {
				RandRand[n+RandOffset] = Buff[n];
				Buff[n] = 0;
				n++;
			}
		}

		/*****************************************************************************
		Funkcja generuje 16-bajtowy blok pseudolosowy
		We:
			Adres na blok do wype-nienia wartociami losowymi
		Wy:
			blok danych pseudolosowych

		*****************************************************************************/
		static public void TVTT_Rand(ref byte[] Rand, int offset)
		{
			int n;
			byte i;
			byte[] NewKeyA = new byte[16];
			byte[] NewKeyB = new byte[16];

			// obsuga mieszania z danymi losowymi

			n = 0;
			while (n<32) {                                                      // przejrzenie wszystkich kolejek pod kontem iloci zdarze w kolejce
				if (RandRandCnt[n]>(1UL<<n)) {                                  // sprawdzenie czy wystarczajco duzo zdarze
					// jesli do zdarze to wymieszanie zdarze z kluczem generacji
					// przekopiowanie aktualnych zdarze i wyczyszczenie kolejki
					i = 0;
					while (i<8) {
						NewKeyA[i] = RandRand[i+(n*16)];
						NewKeyB[i] = RandRand[i+8+(n*16)];
						NewKeyA[i+8] = 0;
						NewKeyB[i+8] = 0;
						RandRand[i+(n*16)] = 0;
						RandRand[i+8+(n*16)] = 0;
						i++;
					}

					// wygenerowanie nowego klucza
					rijndael.rijndaelEncrypt(RandKeyA, ref NewKeyA);
					rijndael.rijndaelEncrypt(RandKeyB, ref NewKeyB);

					// zapisanie nowych kluczy generacji i wyczyszczenie buforw pomocniczych
					i = 0;
					while (i<16) {
						RandKeyA[i] = NewKeyA[i];
						RandKeyB[i] = NewKeyB[i];
						NewKeyA[i] = 0;
						NewKeyB[i] = 0;
						i++;
					}
				}
				n++;
			}

			// przekopiowanie aktualnego stanu licznika
			n = 0;
			while (n<16) {
				Rand[n+offset] = RandCnt[n];
				n++;
			}

			// zwikszenie licznika
			RandCnt[0]++;                                                       // zwikszenie najmodszego bajtu
			n = 1;
			while (n<16) {                                                      // sprawdzenie przeniesienia na starszy bajt
				if (RandCnt[n-1] == 0)                                          // sprawdzenie czy modszy bajt wyzerowany
					RandCnt[n]++;                                               //  wyzerowany - moliwe tylko gdy nastpio przeniesienie -> zwikszenie starszego bajtu
				else
					break;                                                      // nie wyzerowany - nie wystpio przeniesienie -> koniec ptli
				n++;
			}

			// wygenerowanei wartoci wyjciowej (pseudolosowej)
			rijndael.rijndaelEncrypt(RandKeyA, ref Rand);
			rijndael.rijndaelEncrypt(RandKeyB, ref Rand);

			// wygenerowanie nowego pierwszego klucza

			// przekopiowanie aktualnego stanu licznika
			n = 0;
			while (n<16) {
				NewKeyA[n] = RandCnt[n];
				n++;
			}

			// zwikszenie licznika
			RandCnt[0]++;                                                       // zwikszenie najmodszego bajtu
			n = 1;
			while(n<16) {                                                       // sprawdzenie przeniesienia na starszy bajt
				if (RandCnt[n-1] == 0)                                          // sprawdzenie czy modszy bajt wyzerowany
					RandCnt[n]++;                                               //  wyzerowany - moliwe tylko gdy nastpio przeniesienie -> zwikszenie starszego bajtu
				else
					break;                                                      // nie wyzerowany - nie wystpio przeniesienie -> koniec ptli
				n++;
			}

			// wygenerowanie nowego klucza
			rijndael.rijndaelEncrypt(RandKeyA, ref NewKeyA);
			rijndael.rijndaelEncrypt(RandKeyB, ref NewKeyA);

			// wygenerowanie nowego drugiego klucza

			// przekopiowanie aktualnego stanu licznika
			n = 0;
			while (n<16) {
				NewKeyB[n] = RandCnt[n];
				n++;
			}

			// zwikszenie licznika
			RandCnt[0]++;                                                       // zwikszenie najmodszego bajtu
			n = 1;
			while (n<16) {                                                      // sprawdzenie przeniesienia na starszy bajt
				if (RandCnt[n-1] == 0)                                          // sprawdzenie czy modszy bajt wyzerowany
					RandCnt[n]++;                                               // wyzerowany - moliwe tylko gdy nastpio przeniesienie -> zwikszenie starszego bajtu
				else
					break;                                                      // nie wyzerowany - nie wystpio przeniesienie -> koniec ptli
				n++;
			}

			// wygenerowanie nowego klucza
			rijndael.rijndaelEncrypt(RandKeyA, ref NewKeyB);
			rijndael.rijndaelEncrypt(RandKeyB, ref NewKeyB);

			// zapamitanie nowych kluczy i wyczyszczenie buforw pomocniczych
			n = 0;
			while (n<16) {
				RandKeyA[n] = NewKeyA[n];
				RandKeyB[n] = NewKeyB[n];
				NewKeyA[n] = 0;
				NewKeyB[n] = 0;
				n++;
			}
		}

		static public void TVTT_Rand(ref byte[] Rand)
		{
			TVTT_Rand(ref Rand, 0);
		}

		/*****************************************************************************
		 Wrappery funkcji z dll
		*****************************************************************************/

		static public uint GetLastError()
		{
			return LibWrap.GetLastError();
		}

		static public void SetLastError(uint dwErrCode)
		{
			LibWrap.SetLastError(dwErrCode);
		}

		static public uint TVTT_GetVersion(ref uint Ver,
										   ref uint SubVer,
										   ref uint ApiVer)
		{
			return LibWrap.TVTT_GetVersion(ref Ver, ref SubVer, ref ApiVer);
		}

		static public uint TVTT_Talk(uint TokenH,
									 uint TimeOut,
									 uint OutKey,
									 ref byte OutDataA,
									 ref byte OutDataB,
									 ref uint InKey,
									 ref byte InDataA,
									 ref byte InDataB)
		{
			return LibWrap.TVTT_Talk(TokenH, TimeOut, OutKey, ref OutDataA,
				ref OutDataB, ref InKey, ref InDataA, ref InDataB);
		}

		static public uint TVTT_FindToken(ref uint TokenH,
										  uint TimeOut,
										  uint OwnerID,
										  byte OwnerMsk,
										  byte Skip,
										  ref uint FeatureTblSize,
										  ref uint FeatureTbl)
		{
			return LibWrap.TVTT_FindToken(ref TokenH, TimeOut, OwnerID,
				OwnerMsk, Skip, ref FeatureTblSize, ref FeatureTbl);
		}

		static public uint TVTT_StartSession(uint TokenH, uint TimeOut)
		{
			return LibWrap.TVTT_StartSession(TokenH, TimeOut);
		}

		static public uint TVTT_EndSession(uint TokenH)
		{
			return LibWrap.TVTT_EndSession(TokenH);
		}

		static public uint TVTT_FreeToken(uint TokenH)
		{
			return LibWrap.TVTT_FreeToken(TokenH);
		}
	}
}
