using System;

using ExCommon;

namespace TeleVox_TeleToken
{
	public class CTVTTAPI_INT_CON: Object
	{
		private const short TVTTAPI_INT_CON_SIZE = 1104;

		public byte[] Data = new byte[TVTTAPI_INT_CON_SIZE];

		public CTVTTAPI_INT_CON(): base()
		{
		}

		/*****************************************************************************
		Metoda czyszczca tabel z danymi
		*****************************************************************************/
		public void Free()
		{
			ZeroMem.ZeroMemory(ref Data, TVTTAPI_INT_CON_SIZE);
		}

		/*****************************************************************************
		Metoda zwracajca rozmiar wew. struktury danych
		(w celu zachowania kompatybilnoci wstecz)
		*****************************************************************************/
		static public int GetSize()
		{
		  return TVTTAPI_INT_CON_SIZE;
		}

		/*****************************************************************************
		Metoda czyszczca tablic Rand
		*****************************************************************************/
		public void ZeroRand()
		{
		  ZeroMem.ZeroMemory(ref Data, 16, 0);
		}

		/*****************************************************************************
		Metoda czyszczca tablic RandPC
		*****************************************************************************/
		public void ZeroRandPC()
		{
		  ZeroMem.ZeroMemory(ref Data, 16, 16);
		}

		/*****************************************************************************
		Metoda czyszczca tablic RandT
		*****************************************************************************/
		public void ZeroRandT()
		{
		  ZeroMem.ZeroMemory(ref Data, 16, 32);
		}

		/*****************************************************************************
		Metoda czyszczca tablic Key
		*****************************************************************************/
		public void ZeroKey()
		{
		  ZeroMem.ZeroMemory(ref Data, 16, 48);
		}

		/*****************************************************************************
		Metoda kopiujca tablic RandPC z obiektu klasy CTVTTAPI_RandPC
		*****************************************************************************/
		public void CopyRandPC(CTVTTAPI_RandPC PC)
		{
		  ZeroMem.CopyTable(ref Data, PC.Data, 16, 0, 16);
		}

		/*****************************************************************************
		Metoda kopiujca tablic RandT z obiektu klasy CTVTTAPI_RandT
		*****************************************************************************/
		public void CopyRandT(CTVTTAPI_RandT T)
		{
		  ZeroMem.CopyTable(ref Data, T.Data, 32, 0, 16);
		}

		/*****************************************************************************
		Metoda kopiujca tablic Key z obiektu klasy CTVTTAPI_Key
		*****************************************************************************/
		public void CopyKey(CTVTTAPI_Key K)
		{
		  ZeroMem.CopyTable(ref Data, K.Data, 48, 0, 16);
		}

		/*****************************************************************************
		Metoda zwracajca pojedynczy element tablicy Rand
		*****************************************************************************/
		public byte Rand(int index)
		{
			if ((index < 0) || (index > 15)) return 0;
			return Data[index];
		}

		/*****************************************************************************
		Metoda ustawiajca pojedynczy element tablicy Rand
		*****************************************************************************/
		public void Rand(int index, byte value)
		{
			if ((index < 0) || (index > 15)) return;
			Data[index] = value;
		}

		/*****************************************************************************
		Metoda zwracajca pojedynczy element tablicy RandPC
		*****************************************************************************/
		public byte RandPC(int index)
		{
			if ((index < 16) || (index > 31)) return 0;
			return Data[index];
		}

		/*****************************************************************************
		Metoda ustawiajca pojedynczy element tablicy RandPC
		*****************************************************************************/
		public void RandPC(int index, byte value)
		{
			if ((index < 16) || (index > 31)) return;
			Data[index] = value;
		}

		/*****************************************************************************
		Metoda zwracajca pojedynczy element tablicy RandT
		*****************************************************************************/
		public byte RandT(int index)
		{
			if ((index < 32) || (index > 47)) return 0;
			return Data[index];
		}

		/*****************************************************************************
		Metoda ustawiajca pojedynczy element tablicy RandT
		*****************************************************************************/
		public void RandT(int index, byte value)
		{
			if ((index < 32) || (index > 47)) return;
			Data[index] = value;
		}

		/*****************************************************************************
		Metoda zwracajca pojedynczy element tablicy Key
		*****************************************************************************/
		public byte Key(int index)
		{
			if ((index < 48) || (index > 63)) return 0;
			return Data[index];
		}

		/*****************************************************************************
		Metoda ustawiajca pojedynczy element tablicy Key
		*****************************************************************************/
		public void Key(int index, byte value)
		{
			if ((index < 48) || (index > 63)) return;
			Data[index] = value;
		}

		/*****************************************************************************
		Metoda zwracajca uchwyt na TeleToken
		*****************************************************************************/
		public uint hToken()
		{
			return (uint)(Data[64] | (Data[65]<<8) | (Data[66]<<16) |
				(Data[67]<<24));
		}

		/*****************************************************************************
		Metoda ustawiajca uchwyt na TeleToken
		*****************************************************************************/
		public void hToken(uint t)
		{
			Data[64] = (byte) t;
			Data[65] = (byte) (t>>8);
			Data[66] = (byte) (t>>16);
			Data[67] = (byte) (t>>24);
		}

		/*****************************************************************************
		Metoda zwracajca OutKey
		*****************************************************************************/
		public uint OutKey()
		{
			return (uint)(Data[68] | (Data[69]<<8) | (Data[70]<<16) |
				(Data[71]<<24));
		}

		/*****************************************************************************
		Metoda ustawiajca OutKey
		*****************************************************************************/
		public void OutKey(uint t)
		{
			Data[68] = (byte) t;
			Data[69] = (byte) (t>>8);
			Data[70] = (byte) (t>>16);
			Data[71] = (byte) (t>>24);
		}

		/*****************************************************************************
		Metoda zwracajca InKey
		*****************************************************************************/
		public uint InKey()
		{
			return (uint)(Data[72] | (Data[73]<<8) | (Data[74]<<16) |
				(Data[75]<<24));
		}

		/*****************************************************************************
		Metoda ustawiajca InKey
		*****************************************************************************/
		public void InKey(uint t)
		{
			Data[72] = (byte) t;
			Data[73] = (byte) (t>>8);
			Data[74] = (byte) (t>>16);
			Data[75] = (byte) (t>>24);
		}

		/*****************************************************************************
		Metoda zwracajca pojedynczy element tablicy TokenInfo
		*****************************************************************************/
		public uint TokenInfo(int index)
		{
			if ((index < 0) || (index > 255)) return 0;
			int shift = 76+index*4;
			return (uint)(Data[shift] | (Data[shift+1]<<8) |
				(Data[shift+2]<<16) | (Data[shift+3]<<24));
		}

		/*****************************************************************************
		Metoda ustawiajca pojedynczy element tablicy Key
		*****************************************************************************/
		public void TokenInfo(int index, uint value)
		{
			if ((index < 0) || (index > 255)) return;
			int shift = 76+index*4;
			Data[shift] = (byte) value;
			Data[shift+1] = (byte) (value>>8);
			Data[shift+2] = (byte) (value>>16);
			Data[shift+3] = (byte) (value>>24);
		}

		/*****************************************************************************
		Metoda zwracajca TokenInfoSize
		*****************************************************************************/
		public uint TokenInfoSize()
		{
			return (uint)(Data[1100] | (Data[1101]<<8) | (Data[1102]<<16) |
				(Data[1103]<<24));
		}

		/*****************************************************************************
		Metoda ustawiajca TokenInfoSize
		*****************************************************************************/
		public void TokenInfoSize(uint t)
		{
			Data[1100] = (byte) t;
			Data[1101] = (byte) (t>>8);
			Data[1102] = (byte) (t>>16);
			Data[1103] = (byte) (t>>24);
		}
	}

	public class TeleTokenAPI
	{
		private const byte REG_BINARY                  = 3;
		private const byte REG_OPTION_NON_VOLATILE     = 0;
		private const byte REG_OPENED_EXISTING_KEY     = 2;

		private const uint HKEY_CURRENT_USER           = 0x80000001;
		private const uint HKEY_LOCAL_MACHINE          = 0x80000002;
		private const uint HKEY_USERS                  = 0x80000003;
		private const uint HKEY_CURRENT_CONFIG         = 0x80000005;
		private const uint HKEY_DYN_DATA               = 0x80000006;

		private const uint KEY_ALL_ACCESS              = 0xF003F;
		private const uint KEY_CREATE_LINK             = 0x20;
		private const uint KEY_CREATE_SUB_KEY          = 0x4;
		private const uint KEY_ENUMERATE_SUB_KEYS      = 0x8;
		private const uint KEY_EXECUTE                 = 0x20019;
		private const uint KEY_NOTIFY                  = 0x10;
		private const uint KEY_QUERY_VALUE             = 0x1;
		private const uint KEY_READ                    = 0x20019;
		private const uint KEY_SET_VALUE               = 0x2;
		private const uint KEY_WRITE                   = 0x20006;

		static private string TVTTAPI_Int_Name = "Software\\App Data\\Seed";
		static private uint TVTTAPI_Int_LastUpdate = 0;

		public TeleTokenAPI()
		{
		}

		/*****************************************************************************
		Metoda wewntrzna - szyfruje strukture opisujc poczenie
		We:
			Klucz szyfrowania
			Obiekt poczenia
		Wy:
		*****************************************************************************/
		static private void TVTTAPI_Int_CodeCon(byte[] Pass, CTVTTAPI_INT_CON Con)
		{
			uint n, i;

			TeleToken.TVTT_Rand(ref Con.Data);
			n = 16;
			while (n<(((CTVTTAPI_INT_CON.GetSize()+15) / 16)*16)) {
				i = 0;
				while (i<16) {
				  Con.Data[n+i] ^= Con.Data[n+i-16];
				  i++;
				}
				rijndael.rijndaelEncrypt(Pass, ref Con.Data, n);
				n += 16;
			}
		}

		/*****************************************************************************
		Metoda wewntrzna - rozszyfruje strukture opisujc poczenie
		We:
			Klucz szyfrowania
			Obiekt poczenia
		Wy:
		*****************************************************************************/
		static private void TVTTAPI_Int_DecodeCon(byte[] Pass, CTVTTAPI_INT_CON Con)
		{
			uint n, i;

			n = (uint)((((CTVTTAPI_INT_CON.GetSize()+15) / 16)*16)-16);
			while (n>15) {
				rijndael.rijndaelDecrypt(Pass, ref Con.Data, n);
				i = 0;
				while (i<16) {
				  Con.Data[n+i] ^= Con.Data[n+i-16];
				  i++;
				}
				n -= 16;
		  	}
		}

		/*****************************************************************************
		Metoda wewntrzna - na podstawie zwrconego kodu bdu ustawia "LastError"
		We:
			kod bdu
		Wy:
			ustawiony "LastError"
		*****************************************************************************/
		static private void TVTTAPI_Int_SetLastError(uint Res)
		{
			switch (Res&0xFF00) {
				case TeleToken.TVTTS_TOO_MANY_OPEN_TOKENS:
					TeleToken.SetLastError(Err.ERROR_TOO_MANY_OPEN_FILES);
					return;

				case TeleToken.TVTTS_USB_DOES_NOT_EXIST:
					TeleToken.SetLastError(Err.ERROR_UNKNOWN_PORT);
					return;

				case TeleToken.TVTTS_TOKEN_NOT_FOUND:
					TeleToken.SetLastError(Err.ERROR_NOT_FOUND);
					return;

				case TeleToken.TVTTS_TIME_OUT:
					TeleToken.SetLastError(Err.ERROR_TIMEOUT);
					return;

				case TeleToken.TVTTS_EXPIRED:
					TeleToken.SetLastError(Err.ERROR_ACCOUNT_EXPIRED);
					return;

				case TeleToken.TVTTS_ALREADY_OPENED:
					TeleToken.SetLastError(Err.ERROR_ALREADY_INITIALIZED);
					return;

				case TeleToken.TVTTS_DLL_INIT_FAILED:
					TeleToken.SetLastError(Err.ERROR_DLL_NOT_FOUND);
					return;

				case TeleToken.TVTTS_UNKNOWN_ERROR:
				case TeleToken.TVTTS_NULL_PTR:
				case TeleToken.TVTTS_BAD_PARAM:
				case TeleToken.TVTTS_INVALID_HANDLE:
					TeleToken.SetLastError(Err.ERROR_UNEXP_NET_ERR);
					return;
			}

			switch (Res&0xFF) {
				case TeleToken.TVTTS_NO_OPEN:
					TeleToken.SetLastError(Err.ERROR_ACCOUNT_EXPIRED);
					return;

				case TeleToken.TVTTS_EXPIRED:
					TeleToken.SetLastError(Err.ERROR_TIMEOUT);
					return;

				case TeleToken.TVTTS_DAMAGED:
					TeleToken.SetLastError(Err.ERROR_GEN_FAILURE);
					return;

				case TeleToken.TVTTS_NOT_SUPPORTED:
					TeleToken.SetLastError(Err.ERROR_NOT_SUPPORTED);
					return;

				case TeleToken.TVTTS_OUT_OF_RANGE:
					TeleToken.SetLastError(Err.ERROR_INVALID_BLOCK);
					return;

				case TeleToken.TVTTS_PROTECTION_NOT_SUPPORTED:
					TeleToken.SetLastError(Err.ERROR_SET_NOT_FOUND);
					return;

				case TeleToken.TVTTS_PASS_NOT_PERMITTED:
					TeleToken.SetLastError(Err.ERROR_INVALID_PASSWORD);
					return;
     
				case TeleToken.TVTTS_NOT_PERMITTED:
					TeleToken.SetLastError(Err.ERROR_ACCESS_DENIED);
					return;

				case TeleToken.TVTTS_ACCESS_EXPIRED:
					TeleToken.SetLastError(Err.ERROR_COUNTER_TIMEOUT);
					return;
			}

			if (Res != 0)
				TeleToken.SetLastError(Err.ERROR_UNEXP_NET_ERR);
			else
				TeleToken.SetLastError(Err.ERROR_SUCCESS);
		}

		/*****************************************************************************
		Metoda wewntrzna - testuje poprawnoc komunikacji
		We:
			kod bdu
		Wy:
			0 - bez bedw / <> 0 - wystapi bd
			ustawiony "LastError"
		*****************************************************************************/
		static public byte TVTTAPI_Int_CheckError(CTVTTAPI_INT_CON Con,
												  uint Res, uint InKey,
												  byte[] InDataA,
												  byte[] InDataB)
		{
			if (Res != 0) {
				TVTTAPI_Int_SetLastError(Res);
				return 1;
			}
			if (InKey == 0) {
				if (TeleToken.TVTT_Check(InDataA, ref InDataB) != 0)
					TeleToken.SetLastError(Err.ERROR_CRC);
				else {
				  if (InDataB[TeleToken.TVTTD_STATUS] != 0)
					TVTTAPI_Int_SetLastError(InDataB[TeleToken.TVTTD_STATUS]);
				  else
					TeleToken.SetLastError(Err.ERROR_UNEXP_NET_ERR);
				}
				return 1;
			} else {
				if ((Con.InKey() & 0xff000000) > 0) Con.InKey(InKey);
				else {
					if (InKey != Con.InKey()) {
						TeleToken.SetLastError(Err.ERROR_UNEXP_NET_ERR);
						return 1;
					}
				}
			  Con.InKey((Con.InKey()+1) & 0xffffff);
			}

			if (TeleToken.TVTT_Check(InDataA, ref InDataB) != 0) {
				TeleToken.SetLastError(Err.ERROR_CRC);
				return 1;
			}

			if (InDataB[TeleToken.TVTTD_STATUS] != 0) {
				TVTTAPI_Int_SetLastError(InDataB[TeleToken.TVTTD_STATUS]);
				return 1;
			}

		  	return 0;
		}

		/*****************************************************************************
		Metoda inicjalizuje bibliotek
		We:
		Wy:
		*****************************************************************************/
		static public void TVTTAPI_Initialize()
		{
			uint hKeyr = 0, Res = 0, n, i, l;
			byte[] TmpTab = new byte[TeleToken.TVTT_RAND_INIT_TBL_SIZE];
			byte[][] InitTab = new byte[TeleToken.TVTT_RAND_INIT_TBL_SIZE/16][];
			for (n=0; n<(TeleToken.TVTT_RAND_INIT_TBL_SIZE/16); n++)
				InitTab[n] = new byte[16];

			n = 0;
			while (n<TeleToken.TVTT_RAND_INIT_TBL_SIZE) {
				InitTab[n/16][n%16] = (byte) MainDef.GetTickCount();
				n++;
			}

			l = LibWrap.RegCreateKeyEx(HKEY_CURRENT_USER, TVTTAPI_Int_Name, 0,
				null, REG_OPTION_NON_VOLATILE, KEY_QUERY_VALUE | KEY_SET_VALUE,
				0, ref hKeyr, ref Res);
			if ((l == Err.ERROR_SUCCESS) || (Res == REG_OPENED_EXISTING_KEY)) {
				n = TeleToken.TVTT_RAND_INIT_TBL_SIZE;
				i = REG_BINARY;
				LibWrap.RegQueryValueEx(hKeyr, "", 0, ref i, ref TmpTab[0],
					ref n);
				for (n=0; n<TeleToken.TVTT_RAND_INIT_TBL_SIZE; n++) {
					InitTab[n/16][n%16] = TmpTab[n];
					TmpTab[n] = 0;
				}
			}

			TeleToken.TVTT_RandInit(ref InitTab);
			TeleToken.TVTT_RandGetInit(ref InitTab);

			for (n=0; n<TeleToken.TVTT_RAND_INIT_TBL_SIZE; n++) {
				TmpTab[n] = InitTab[n/16][n%16];
				InitTab[n/16][n%16] = 0;
			}
			LibWrap.RegSetValueEx(hKeyr, null, 0, REG_BINARY, TmpTab,
				TeleToken.TVTT_RAND_INIT_TBL_SIZE);

			if (l == Err.ERROR_SUCCESS) LibWrap.RegCloseKey(hKeyr);

			ZeroMem.ZeroMemory(ref TmpTab, TeleToken.TVTT_RAND_INIT_TBL_SIZE);
			TeleToken.SetLastError(Err.ERROR_SUCCESS);
		}

		/*****************************************************************************
		Metoda dodaje zdarzenia do generatora liczb pseudolosowych
		We:
			Struktura wiadomoci systemowej
		Wy:
		*****************************************************************************/
		static public void TVTTAPI_Randomize(System.Windows.Forms.Message Msg)
		{
			byte n;
			uint Timer, l, hKeyr = 0, Res = 0;
			byte[][] InitTab = new byte[TeleToken.TVTT_RAND_INIT_TBL_SIZE/16][];
			for (n=0; n<(TeleToken.TVTT_RAND_INIT_TBL_SIZE/16); n++)
				InitTab[n] = new byte[16];

			switch (Msg.Msg) {
				case Msgs.WM_KEYUP: n = 0; break;
				case Msgs.WM_KEYDOWN: n = 1; break;
				case Msgs.WM_MOUSEMOVE: n = 2; break;
				case Msgs.WM_LBUTTONDOWN: n = 3; break;
				case Msgs.WM_LBUTTONUP: n = 4; break;
				case Msgs.WM_MBUTTONDOWN: n = 5; break;
				case Msgs.WM_MBUTTONUP: n = 6; break;
				case Msgs.WM_RBUTTONDOWN: n = 7; break;
				case Msgs.WM_RBUTTONUP: n = 8; break;
			}

			if ((n*3)<31) {
				TeleToken.TVTT_Randomize((byte)((n*3)+0),
					MainDef.GetTickCount());
				TeleToken.TVTT_Randomize((byte)((n*3)+1),
					(uint)(Msg.WParam.ToInt32()));
				TeleToken.TVTT_Randomize((byte)((n*3)+2),
					(uint)(Msg.LParam.ToInt32()));
			}

			Timer = MainDef.GetTickCount();
			if ((TVTTAPI_Int_LastUpdate & 0x80000000) > 0) {
				if ((Timer & 0x80000000) > 0) {
					if ((TVTTAPI_Int_LastUpdate+10000) < TVTTAPI_Int_LastUpdate)
						return;
					if (MainDef.Time < (TVTTAPI_Int_LastUpdate+10000)) return;
				}
			} else if (MainDef.Time < (TVTTAPI_Int_LastUpdate+10000)) return;

			TVTTAPI_Int_LastUpdate = Timer;

			l = LibWrap.RegCreateKeyEx(HKEY_CURRENT_USER, TVTTAPI_Int_Name, 0,
				"", REG_OPTION_NON_VOLATILE, KEY_QUERY_VALUE | KEY_SET_VALUE,
				0, ref hKeyr, ref Res);
			if ((l == Err.ERROR_SUCCESS) && (Res == REG_OPENED_EXISTING_KEY)) {
				TeleToken.TVTT_RandGetInit(ref InitTab);

				byte[] TmpTab = new byte[TeleToken.TVTT_RAND_INIT_TBL_SIZE];
				for (n=0; n<TeleToken.TVTT_RAND_INIT_TBL_SIZE; n++) {
					TmpTab[n] = InitTab[n/16][n%16];
					InitTab[n/16][n%16] = 0;
				}
				LibWrap.RegSetValueEx(hKeyr, "", 0, REG_BINARY, TmpTab,
					TeleToken.TVTT_RAND_INIT_TBL_SIZE);
				ZeroMem.ZeroMemory(ref TmpTab, TeleToken.TVTT_RAND_INIT_TBL_SIZE);
			}
			if (l == Err.ERROR_SUCCESS) LibWrap.RegCloseKey(hKeyr);
		}

		/*****************************************************************************
		Metoda wypenia podany blok wartoci pseudolosow
		We:
			Adres na tablice do wypenienia
		Wy:
		*****************************************************************************/
		static public void TVTTAPI_Rand(ref byte[] Random)
		{
		  if (Random == null) {
			TeleToken.SetLastError(Err.ERROR_INVALID_DATA);
			return;
		  }
		  TeleToken.TVTT_Rand(ref Random);
		  TeleToken.SetLastError(Err.ERROR_SUCCESS);
		}

		/*****************************************************************************
		Metoda znajduje Token i otwiera poczenie z Tokenem
		We:
			Adres na 16-bajtowy klucz do poczenia
			ID odbiorcy tokenu
			Maska odbiorcy tokenu
			Ilo Tokenw do pominicia
		Wy:
			obiekt poczenia z tokenem lub nil jeli bd
			ustawiony "LastError"
		*****************************************************************************/
		static public CTVTTAPI_INT_CON TVTTAPI_Open(byte[] Pass,
													uint OwnerID,
													byte OwnerMsk, byte Skip)
		{
			if (Pass == null) {
				TeleToken.SetLastError(Err.ERROR_INVALID_DATA);
				return null;
			}

			CTVTTAPI_INT_CON Con = new CTVTTAPI_INT_CON();

			Con.TokenInfoSize(256);
			uint t = 0, tis = Con.TokenInfoSize();
			uint[] ti = new uint[Con.TokenInfoSize()];
			uint Res = TeleToken.TVTT_FindToken(ref t, 5*10, OwnerID, OwnerMsk,
				Skip, ref tis, ref ti[0]);
			if (Res != 0) {
				TVTTAPI_Int_SetLastError(Res);
				Con.Free();
				Con = null;
				return null;
			}
			Con.hToken(t);
			Con.TokenInfoSize(tis);
			for (t=0; t<tis; t++) Con.TokenInfo((int) t, ti[t]);
			TVTTAPI_Int_CodeCon(Pass, Con);
			TeleToken.SetLastError(Err.ERROR_SUCCESS);
			return Con;
		}

		/*****************************************************************************
		Metoda zamyka poczenie z Tokenem i zwalnia zasoby
		We:
			Adres na 16-bajtowy klucz do poczenia
			Obiekt poczenia
		Wy:
			ustawiony "LastError"
		*****************************************************************************/
		static public void TVTTAPI_Close(byte[] Pass, CTVTTAPI_INT_CON Con)
		{
			uint Res;

			if ((Pass == null) || (Con == null)) {
				TeleToken.SetLastError(Err.ERROR_INVALID_DATA);
				return;
			}

			TVTTAPI_Int_DecodeCon(Pass, Con);

			Res = TeleToken.TVTT_FreeToken(Con.hToken());
			if (Res != 0) {
				TVTTAPI_Int_CodeCon(Pass, Con);
				TVTTAPI_Int_SetLastError(Res);
				return;
			}

			Con.Free();
			Con = null;
			TeleToken.SetLastError(Err.ERROR_SUCCESS);
		}

		/*****************************************************************************
		Metoda zwraca warto opisujc konkretn waciwo Tokenu
		We:
			Adres na 16-bajtowy klucz do poczenia
			Obiekt poczenia
			Numer waciwoci Tokenu
		Wy:
			waciwo Tokenu
			ustawiony "LastError"
		*****************************************************************************/
		static public uint TVTTAPI_GetFeature(byte[] Pass, CTVTTAPI_INT_CON Con,
											  uint FNr)
		{
			if ((Pass == null) || (Con == null)) {
				TeleToken.SetLastError(Err.ERROR_INVALID_DATA);
				return 0;
			}

			TVTTAPI_Int_DecodeCon(Pass, Con);

			uint Res = 0;
			if (FNr<Con.TokenInfoSize()) Res = Con.TokenInfo((int) FNr);

			TVTTAPI_Int_CodeCon(Pass, Con);

			TeleToken.SetLastError(Err.ERROR_SUCCESS);
			return Res;
		}

		/*****************************************************************************
		Metoda rozpoczyna sesj z tokenem
		We:
			Adres na 16-bajtowy klucz do poczenia
			Obiekt poczenia
			Numer hasa proponowanego przez PC
		Wy:
			Numer hasa zaproponowanego przez Token
			ustawiony "LastError"
		*****************************************************************************/
		static public byte TVTTAPI_StartSession(byte[] Pass,
												CTVTTAPI_INT_CON Con,
												byte PCPassNr)
		{
			uint Res, InKey = 0;
			byte TPassNr;
			byte[] OutData = new byte[16];
			byte[] InData = new byte[16];

			if ((Pass == null) || (Con == null)) {
				TeleToken.SetLastError(Err.ERROR_INVALID_DATA);
				return 0;
			}

			TVTTAPI_Int_DecodeCon(Pass, Con);

			Res = TeleToken.TVTT_StartSession(Con.hToken(), 1);
			if (Res != 0) {
				TVTTAPI_Int_CodeCon(Pass,Con);
				TVTTAPI_Int_SetLastError(Res);
				return 0;
			}

			Con.ZeroRandPC();
			ZeroMem.ZeroMemory(ref OutData, 16);
			Con.ZeroRandT();
			ZeroMem.ZeroMemory(ref InData, 16);

			Con.OutKey(0);
			Con.InKey(0xffffffff);

			CTVTTAPI_RandPC RandPC = new CTVTTAPI_RandPC();
			TeleToken.TVTT_Rand(ref RandPC.Data);


			OutData[TeleToken.TVTTD_CMD] = TeleToken.TVTTC_OPEN;
			OutData[TeleToken.TVTTD_PARAM1] = PCPassNr;

			CTVTTAPI_RandT RandT = new CTVTTAPI_RandT();
			TeleToken.TVTT_Calc(RandPC.Data, ref OutData);
			Res = TeleToken.TVTT_Talk(Con.hToken(), 5, Con.OutKey(),
				ref RandPC.Data[0], ref OutData[0], ref InKey,
				ref RandT.Data[0], ref InData[0]);
			Con.CopyRandPC(RandPC);
			Con.CopyRandT(RandT);
			RandPC.Free();
			RandPC = null;
			RandT.Free();
			RandT = null;

			if (Res != 0) {
				InKey = 0;
				Con.ZeroRandPC();
				ZeroMem.ZeroMemory(ref OutData, 16);
				Con.ZeroRandT();
				ZeroMem.ZeroMemory(ref InData, 16);
				TeleToken.TVTT_EndSession(Con.hToken());
				TVTTAPI_Int_CodeCon(Pass, Con);
				TVTTAPI_Int_SetLastError(Res);
				return 0;
			}

			if (InKey != 0) {
				InKey = 0;
				Con.ZeroRandPC();
				ZeroMem.ZeroMemory(ref OutData, 16);
				Con.ZeroRandT();
				ZeroMem.ZeroMemory(ref InData, 16);
				TeleToken.TVTT_EndSession(Con.hToken());
				TVTTAPI_Int_CodeCon(Pass, Con);
				TeleToken.SetLastError(Err.ERROR_UNEXP_NET_ERR);
				return 0;
			}

			RandT = new CTVTTAPI_RandT(Con);
			if (TeleToken.TVTT_Check(RandT.Data, ref InData) != 0) {
				RandT.Free();
				RandT = null;
				InKey = 0;
				Con.ZeroRandPC();
				ZeroMem.ZeroMemory(ref OutData, 16);
				Con.ZeroRandT();
				ZeroMem.ZeroMemory(ref InData, 16);
				TeleToken.TVTT_EndSession(Con.hToken());
				TVTTAPI_Int_CodeCon(Pass, Con);
				TeleToken.SetLastError(Err.ERROR_CRC);
				return 0;
			}
			RandT.Free();
			RandT = null;

			if (InData[TeleToken.TVTTD_STATUS] != 0) {
				TeleToken.TVTT_EndSession(Con.hToken());
				TVTTAPI_Int_CodeCon(Pass, Con);
				TVTTAPI_Int_SetLastError(InData[TeleToken.TVTTD_STATUS]);
				InKey = 0;
				Con.ZeroRandPC();
				ZeroMem.ZeroMemory(ref OutData, 16);
				Con.ZeroRandT();
				ZeroMem.ZeroMemory(ref InData, 16);
			}

			TPassNr = InData[TeleToken.TVTTD_PARAM1];

			ZeroMem.ZeroMemory(ref OutData, 16);
			ZeroMem.ZeroMemory(ref InData, 16);
			InKey = 0;

			TeleToken.TVTT_Rand(ref OutData);

			Con.OutKey(OutData[0]);
			Con.OutKey(Con.OutKey() << 8);
			Con.OutKey(Con.OutKey() | OutData[1]);
			Con.OutKey(Con.OutKey() << 8);
			Con.OutKey(Con.OutKey() | OutData[2]);
			Con.OutKey(Con.OutKey() << 8);
			ZeroMem.ZeroMemory(ref OutData, 16);

			TVTTAPI_Int_CodeCon(Pass, Con);

			TeleToken.SetLastError(Err.ERROR_SUCCESS);
			return TPassNr;
		}

		/****************************************************************************
		Metoda koczy sesj z tokenem
		We:
			Adres na 16-bajtowy klucz do poczenia
			Obiekt poczenia
		Wy:
			ustawiony "LastError"
		*****************************************************************************/
		static public void TVTTAPI_EndSession(byte[] Pass, CTVTTAPI_INT_CON Con)
		{
			uint Res, Res2, InKey = 0;
			byte[] OutDataA = new byte[16];
			byte[] OutDataB = new byte[16];
			byte[] InDataA = new byte[16];
			byte[] InDataB = new byte[16];

			if ((Pass == null) || (Con == null)) {
				TeleToken.SetLastError(Err.ERROR_INVALID_DATA);
				return;
			}

			TVTTAPI_Int_DecodeCon(Pass, Con);

			ZeroMem.ZeroMemory(ref OutDataA, 16);
			ZeroMem.ZeroMemory(ref OutDataB, 16);
			ZeroMem.ZeroMemory(ref InDataA, 16);
			ZeroMem.ZeroMemory(ref InDataB, 16);

			Con.OutKey(Con.OutKey() + 1);
			OutDataB[TeleToken.TVTTD_CMD] = TeleToken.TVTTC_CLOSE;

			TeleToken.TVTT_Calc(OutDataA, ref OutDataB);
			CTVTTAPI_Key Key = new CTVTTAPI_Key(Con);
			TeleToken.TVTT_CryptBlock(ref OutDataA, Con.OutKey(), Key.Data,
				TeleToken.TVTT_BLOCK_A, TeleToken.TVTT_PC2T);
			TeleToken.TVTT_CryptBlock(ref OutDataB, Con.OutKey(), Key.Data,
				TeleToken.TVTT_BLOCK_B, TeleToken.TVTT_PC2T);
			Res = TeleToken.TVTT_Talk(Con.hToken(), 5, Con.OutKey(),
				ref OutDataA[0], ref OutDataB[0], ref InKey,
				ref InDataA[0], ref InDataB[0]);
			Key.Free();
			Key = null;
			if (Res != 0) TVTTAPI_Int_SetLastError(Res);

			Res2 = TeleToken.TVTT_EndSession(Con.hToken());
			if ((Res == 0) && (Res2 != 0)) TVTTAPI_Int_SetLastError(Res2);

			Con.OutKey(0);

			TVTTAPI_Int_CodeCon(Pass, Con);

			ZeroMem.ZeroMemory(ref OutDataA, 16);
			ZeroMem.ZeroMemory(ref OutDataB, 16);
			ZeroMem.ZeroMemory(ref InDataA, 16);
			ZeroMem.ZeroMemory(ref InDataB, 16);
			InKey = 0;

			TeleToken.SetLastError(Err.ERROR_SUCCESS);
		}

		/*****************************************************************************
		Metoda wylicza haso sesji na podstawie zaproponowanych hase
		We:
			Adres na 16-bajtowy klucz do poczenia
			Obiekt poczenia
			Haso proponowane przez PC
			Haso proponowane przez T
		Wy:
			ustawiony "LastError"
		*****************************************************************************/
		static public void TVTTAPI_CalcPass(byte[] Pass, CTVTTAPI_INT_CON Con,
											byte[] PassPC, byte[] PassT)
		{
			if ((Pass == null) || (Con == null) || (PassPC == null) ||
				(PassT == null)) {
				TeleToken.SetLastError(Err.ERROR_INVALID_DATA);
				return;
			}

			TVTTAPI_Int_DecodeCon(Pass, Con);

			CTVTTAPI_Key Key = new CTVTTAPI_Key();
			CTVTTAPI_RandPC RandPC = new CTVTTAPI_RandPC(Con);
			CTVTTAPI_RandT RandT = new CTVTTAPI_RandT(Con);
			TeleToken.TVTT_CalcSessionKey(ref Key.Data, PassPC, RandPC.Data, PassT, RandT.Data);
			Con.CopyKey(Key);
			Key.Free();
			RandPC.Free();
			RandT.Free();

			Con.ZeroRandPC();
			Con.ZeroRandT();

			TVTTAPI_Int_CodeCon(Pass, Con);

			TeleToken.SetLastError(Err.ERROR_SUCCESS);
		}

		/*****************************************************************************
		Metoda odczytuje sektor z pamici nieulotnej
		We:
			Adres na 16-bajtowy klucz do poczenia
			Obiekt poczenia
			Numer sektora do odczytania
			Tablica bufora na sektor
		Wy:
			wypeniony bufor na sektor (var Buff)
			ustawiony "LastError"
		*****************************************************************************/
		static public void TVTTAPI_MemRead(byte[] Pass, CTVTTAPI_INT_CON Con,
										   byte SNr, ref byte[] Buff)
		{
			uint Res, InKey = 0;
			byte[] OutDataA = new byte[16];
			byte[] OutDataB = new byte[16];
			byte[] InDataA = new byte[16];
			byte[] InDataB = new byte[16];

			if ((Pass == null) || (Con == null) || (Buff == null)) {
				TeleToken.SetLastError(Err.ERROR_INVALID_DATA);
				return;
			}

			ZeroMem.ZeroMemory(ref Buff, 16);

			TVTTAPI_Int_DecodeCon(Pass, Con);

			ZeroMem.ZeroMemory(ref OutDataA, 16);
			ZeroMem.ZeroMemory(ref OutDataB, 16);
			ZeroMem.ZeroMemory(ref InDataA, 16);
			ZeroMem.ZeroMemory(ref InDataB, 16);

			Con.OutKey(Con.OutKey()+1);
			OutDataB[TeleToken.TVTTD_CMD] = TeleToken.TVTTC_NVMEM_READ;
			OutDataB[TeleToken.TVTTD_PARAM1] = SNr;

			TeleToken.TVTT_Calc(OutDataA, ref OutDataB);
			CTVTTAPI_Key Key = new CTVTTAPI_Key(Con);
			TeleToken.TVTT_CryptBlock(ref OutDataA, Con.OutKey(), Key.Data,
				TeleToken.TVTT_BLOCK_A, TeleToken.TVTT_PC2T);
			TeleToken.TVTT_CryptBlock(ref OutDataB, Con.OutKey(), Key.Data,
				TeleToken.TVTT_BLOCK_B, TeleToken.TVTT_PC2T);
			Res = TeleToken.TVTT_Talk(Con.hToken(), 5, Con.OutKey(),
				ref OutDataA[0], ref OutDataB[0], ref InKey,
				ref InDataA[0], ref InDataB[0]);
			TeleToken.TVTT_CryptBlock(ref InDataA, InKey, Key.Data,
				TeleToken.TVTT_BLOCK_A, TeleToken.TVTT_T2PC);
			TeleToken.TVTT_CryptBlock(ref InDataB, InKey, Key.Data,
				TeleToken.TVTT_BLOCK_B, TeleToken.TVTT_T2PC);
			Key.Free();
			if (TVTTAPI_Int_CheckError(Con, Res, InKey, InDataA, InDataB) != 0) {
				InKey = 0;
				ZeroMem.ZeroMemory(ref OutDataA, 16);
				ZeroMem.ZeroMemory(ref OutDataB, 16);
				ZeroMem.ZeroMemory(ref InDataA, 16);
				ZeroMem.ZeroMemory(ref InDataB, 16);
				TVTTAPI_Int_CodeCon(Pass, Con);
				return;
			}

			ZeroMem.CopyTable(ref Buff, InDataA, 0, 0, 16);

			TVTTAPI_Int_CodeCon(Pass, Con);

			InKey = 0;
			ZeroMem.ZeroMemory(ref OutDataA, 16);
			ZeroMem.ZeroMemory(ref OutDataB, 16);
			ZeroMem.ZeroMemory(ref InDataA, 16);
			ZeroMem.ZeroMemory(ref InDataB, 16);

			TeleToken.SetLastError(Err.ERROR_SUCCESS);
		}

		/*****************************************************************************
		Metoda odczytuje sektor z pamici ulotnej
		We:
			Adres na 16-bajtowy klucz do poczenia
			Obiekt poczenia
			Numer sektora do odczytania
			Tablica bufora na sektor
		Wy:
			wypeniony bufor na sektor (var Buff)
			ustawiony "LastError"
		*****************************************************************************/
		static public void TVTTAPI_RamRead(byte[] Pass, CTVTTAPI_INT_CON Con,
										   byte SNr, ref byte[] Buff)
		{
			uint Res, InKey = 0;
			byte[] OutDataA = new byte[16];
			byte[] OutDataB = new byte[16];
			byte[] InDataA = new byte[16];
			byte[] InDataB = new byte[16];

			if ((Pass == null) || (Con == null) || (Buff == null)) {
				TeleToken.SetLastError(Err.ERROR_INVALID_DATA);
				return;
			}

			ZeroMem.ZeroMemory(ref Buff, 16);

			TVTTAPI_Int_DecodeCon(Pass, Con);

			ZeroMem.ZeroMemory(ref OutDataA, 16);
			ZeroMem.ZeroMemory(ref OutDataB, 16);
			ZeroMem.ZeroMemory(ref InDataA, 16);
			ZeroMem.ZeroMemory(ref InDataB, 16);


			Con.OutKey(Con.OutKey()+1);
			OutDataB[TeleToken.TVTTD_CMD] = TeleToken.TVTTC_VMEM_READ;
			OutDataB[TeleToken.TVTTD_PARAM1] = SNr;

			TeleToken.TVTT_Calc(OutDataA, ref OutDataB);
			CTVTTAPI_Key Key = new CTVTTAPI_Key(Con);
			TeleToken.TVTT_CryptBlock(ref OutDataA, Con.OutKey(), Key.Data,
				TeleToken.TVTT_BLOCK_A, TeleToken.TVTT_PC2T);
			TeleToken.TVTT_CryptBlock(ref OutDataB, Con.OutKey(), Key.Data,
				TeleToken.TVTT_BLOCK_B, TeleToken.TVTT_PC2T);
			Res = TeleToken.TVTT_Talk(Con.hToken(), 5, Con.OutKey(),
				ref OutDataA[0], ref OutDataB[0], ref InKey,
				ref InDataA[0], ref InDataB[0]);
			TeleToken.TVTT_CryptBlock(ref InDataA, InKey, Key.Data,
				TeleToken.TVTT_BLOCK_A, TeleToken.TVTT_T2PC);
			TeleToken.TVTT_CryptBlock(ref InDataB, InKey, Key.Data,
				TeleToken.TVTT_BLOCK_B, TeleToken.TVTT_T2PC);
			Key.Free();
			if (TVTTAPI_Int_CheckError(Con, Res, InKey, InDataA, InDataB) != 0) {
				InKey = 0;
				ZeroMem.ZeroMemory(ref OutDataA, 16);
				ZeroMem.ZeroMemory(ref OutDataB, 16);
				ZeroMem.ZeroMemory(ref InDataA, 16);
				ZeroMem.ZeroMemory(ref InDataB, 16);
				TVTTAPI_Int_CodeCon(Pass, Con);
				return;
			}

			ZeroMem.CopyTable(ref Buff, InDataA, 0, 0, 16);

			TVTTAPI_Int_CodeCon(Pass, Con);

			InKey = 0;
			ZeroMem.ZeroMemory(ref OutDataA, 16);
			ZeroMem.ZeroMemory(ref OutDataB, 16);
			ZeroMem.ZeroMemory(ref InDataA, 16);
			ZeroMem.ZeroMemory(ref InDataB, 16);

		  	TeleToken.SetLastError(Err.ERROR_SUCCESS);
		}

		/*****************************************************************************
		Metoda zapisuje sektor do pamici ulotnej
		We:
			Adres na 16-bajtowy klucz do poczenia
			Obiekt poczenia
			Numer sektora do zapisania
			Tablica bufora z danymi do zapisania
		Wy:
			ustawiony "LastError"
		*****************************************************************************/
		static public void TVTTAPI_RamWrite(byte[] Pass, CTVTTAPI_INT_CON Con,
											byte SNr, byte[] Buff)
		{
			uint Res, InKey = 0;
			byte[] OutDataA = new byte[16];
			byte[] OutDataB = new byte[16];
			byte[] InDataA = new byte[16];
			byte[] InDataB = new byte[16];

			if ((Pass == null) || (Con == null) || (Buff == null)) {
				TeleToken.SetLastError(Err.ERROR_INVALID_DATA);
				return;
			}

			TVTTAPI_Int_DecodeCon(Pass, Con);

			ZeroMem.CopyTable(ref OutDataA, Buff, 0, 0, 16);
			ZeroMem.ZeroMemory(ref OutDataB, 16);
			ZeroMem.ZeroMemory(ref InDataA, 16);
			ZeroMem.ZeroMemory(ref InDataB, 16);

			Con.OutKey(Con.OutKey()+1);
			OutDataB[TeleToken.TVTTD_CMD] = TeleToken.TVTTC_VMEM_WRITE;
			OutDataB[TeleToken.TVTTD_PARAM1] = SNr;

			TeleToken.TVTT_Calc(OutDataA, ref OutDataB);
			CTVTTAPI_Key Key = new CTVTTAPI_Key(Con);
			TeleToken.TVTT_CryptBlock(ref OutDataA, Con.OutKey(), Key.Data,
				TeleToken.TVTT_BLOCK_A, TeleToken.TVTT_PC2T);
			TeleToken.TVTT_CryptBlock(ref OutDataB, Con.OutKey(), Key.Data,
				TeleToken.TVTT_BLOCK_B, TeleToken.TVTT_PC2T);
			Res = TeleToken.TVTT_Talk(Con.hToken(), 5, Con.OutKey(),
				ref OutDataA[0], ref OutDataB[0], ref InKey,
				ref InDataA[0], ref InDataB[0]);
			TeleToken.TVTT_CryptBlock(ref InDataA, InKey, Key.Data,
				TeleToken.TVTT_BLOCK_A, TeleToken.TVTT_T2PC);
			TeleToken.TVTT_CryptBlock(ref InDataB, InKey, Key.Data,
				TeleToken.TVTT_BLOCK_B, TeleToken.TVTT_T2PC);
			Key.Free();
			if (TVTTAPI_Int_CheckError(Con, Res, InKey, InDataA, InDataB) != 0) {
				InKey = 0;
				ZeroMem.ZeroMemory(ref OutDataA, 16);
				ZeroMem.ZeroMemory(ref OutDataB, 16);
				ZeroMem.ZeroMemory(ref InDataA, 16);
				ZeroMem.ZeroMemory(ref InDataB, 16);
				TVTTAPI_Int_CodeCon(Pass, Con);
				return;
			}

			TVTTAPI_Int_CodeCon(Pass, Con);

			InKey = 0;
			ZeroMem.ZeroMemory(ref OutDataA, 16);
			ZeroMem.ZeroMemory(ref OutDataB, 16);
			ZeroMem.ZeroMemory(ref InDataA, 16);
			ZeroMem.ZeroMemory(ref InDataB, 16);

			TeleToken.SetLastError(Err.ERROR_SUCCESS);
		}

		/*****************************************************************************
		Metoda zapisuje sektor do pamici nieulotnej
		We:
			Adres na 16-bajtowy klucz do poczenia
			Obiekt poczenia
			Numer sektora do zapisania
			Tablica bufora z danymi do zapisania
		Wy:
			ustawiony "LastError"
		*****************************************************************************/
		static public void TVTTAPI_MemWrite(byte[] Pass, CTVTTAPI_INT_CON Con,
											byte SNr, byte[] Buff)
		{
			uint Res, InKey = 0;
			byte[] OutDataA = new byte[16];
			byte[] OutDataB = new byte[16];
			byte[] InDataA = new byte[16];
			byte[] InDataB = new byte[16];

			if ((Pass == null) || (Con == null) || (Buff == null)) {
				TeleToken.SetLastError(Err.ERROR_INVALID_DATA);
				return;
			}

			TVTTAPI_Int_DecodeCon(Pass, Con);

			ZeroMem.CopyTable(ref OutDataA, Buff, 0, 0, 16);
			ZeroMem.ZeroMemory(ref OutDataB, 16);
			ZeroMem.ZeroMemory(ref InDataA, 16);
			ZeroMem.ZeroMemory(ref InDataB, 16);

			Con.OutKey(Con.OutKey()+1);
			OutDataB[TeleToken.TVTTD_CMD] = TeleToken.TVTTC_NVMEM_WRITE;
			OutDataB[TeleToken.TVTTD_PARAM1] = SNr;

			TeleToken.TVTT_Calc(OutDataA, ref OutDataB);
			CTVTTAPI_Key Key = new CTVTTAPI_Key(Con);
			TeleToken.TVTT_CryptBlock(ref OutDataA, Con.OutKey(), Key.Data,
				TeleToken.TVTT_BLOCK_A, TeleToken.TVTT_PC2T);
			TeleToken.TVTT_CryptBlock(ref OutDataB, Con.OutKey(), Key.Data,
				TeleToken.TVTT_BLOCK_B, TeleToken.TVTT_PC2T);
			Res = TeleToken.TVTT_Talk(Con.hToken(), 5, Con.OutKey(),
				ref OutDataA[0], ref OutDataB[0], ref InKey,
				ref InDataA[0], ref InDataB[0]);
			TeleToken.TVTT_CryptBlock(ref InDataA, InKey, Key.Data,
				TeleToken.TVTT_BLOCK_A, TeleToken.TVTT_T2PC);
			TeleToken.TVTT_CryptBlock(ref InDataB, InKey, Key.Data,
				TeleToken.TVTT_BLOCK_B, TeleToken.TVTT_T2PC);
			Key.Free();
			if (TVTTAPI_Int_CheckError(Con, Res, InKey, InDataA, InDataB) != 0) {
				InKey = 0;
				ZeroMem.ZeroMemory(ref OutDataA, 16);
				ZeroMem.ZeroMemory(ref OutDataB, 16);
				ZeroMem.ZeroMemory(ref InDataA, 16);
				ZeroMem.ZeroMemory(ref InDataB, 16);
				TVTTAPI_Int_CodeCon(Pass, Con);
				return;
			}

			TVTTAPI_Int_CodeCon(Pass, Con);

			InKey = 0;
			ZeroMem.ZeroMemory(ref OutDataA, 16);
			ZeroMem.ZeroMemory(ref OutDataB, 16);
			ZeroMem.ZeroMemory(ref InDataA, 16);
			ZeroMem.ZeroMemory(ref InDataB, 16);

		  	TeleToken.SetLastError(Err.ERROR_SUCCESS);
		}

		/*****************************************************************************
		Metoda zapisuje sektor do pamici nieulotnej i zabezpiecza sektor przed zapisem
		We:
			Adres na 16-bajtowy klucz do poczenia
			Obiekt poczenia
			Numer sektora do zapisania
			Tablica bufora z danymi do zapisania
		Wy:
			ustawiony "LastError"
		*****************************************************************************/
		static public void TVTTAPI_MemWriteProtect(byte[] Pass,
												   CTVTTAPI_INT_CON Con,
												   byte SNr, byte[] Buff)
		{
			uint Res, InKey = 0;
			byte[] OutDataA = new byte[16];
			byte[] OutDataB = new byte[16];
			byte[] InDataA = new byte[16];
			byte[] InDataB = new byte[16];

			if ((Pass == null) || (Con == null) || (Buff == null)) {
				TeleToken.SetLastError(Err.ERROR_INVALID_DATA);
				return;
			}

			TVTTAPI_Int_DecodeCon(Pass, Con);

			ZeroMem.CopyTable(ref OutDataA, Buff, 0, 0, 16);
			ZeroMem.ZeroMemory(ref OutDataB, 16);
			ZeroMem.ZeroMemory(ref InDataA, 16);
			ZeroMem.ZeroMemory(ref InDataB, 16);

			Con.OutKey(Con.OutKey()+1);
			OutDataB[TeleToken.TVTTD_CMD] = TeleToken.TVTTC_NVMEM_WRITE_PROTECT;
			OutDataB[TeleToken.TVTTD_PARAM1] = SNr;

			TeleToken.TVTT_Calc(OutDataA, ref OutDataB);
			CTVTTAPI_Key Key = new CTVTTAPI_Key(Con);
			TeleToken.TVTT_CryptBlock(ref OutDataA, Con.OutKey(), Key.Data,
				TeleToken.TVTT_BLOCK_A, TeleToken.TVTT_PC2T);
			TeleToken.TVTT_CryptBlock(ref OutDataB, Con.OutKey(), Key.Data,
				TeleToken.TVTT_BLOCK_B, TeleToken.TVTT_PC2T);
			Res = TeleToken.TVTT_Talk(Con.hToken(), 10, Con.OutKey(),
				ref OutDataA[0], ref OutDataB[0], ref InKey,
				ref InDataA[0], ref InDataB[0]);
			TeleToken.TVTT_CryptBlock(ref InDataA, InKey, Key.Data,
				TeleToken.TVTT_BLOCK_A, TeleToken.TVTT_T2PC);
			TeleToken.TVTT_CryptBlock(ref InDataB, InKey, Key.Data,
				TeleToken.TVTT_BLOCK_B, TeleToken.TVTT_T2PC);
			Key.Free();
			if (TVTTAPI_Int_CheckError(Con, Res, InKey, InDataA, InDataB) != 0) {
				InKey = 0;
				ZeroMem.ZeroMemory(ref OutDataA, 16);
				ZeroMem.ZeroMemory(ref OutDataB, 16);
				ZeroMem.ZeroMemory(ref InDataA, 16);
				ZeroMem.ZeroMemory(ref InDataB, 16);
				TVTTAPI_Int_CodeCon(Pass, Con);
				return;
			}

			TVTTAPI_Int_CodeCon(Pass, Con);

			InKey = 0;
			ZeroMem.ZeroMemory(ref OutDataA, 16);
			ZeroMem.ZeroMemory(ref OutDataB, 16);
			ZeroMem.ZeroMemory(ref InDataA, 16);
			ZeroMem.ZeroMemory(ref InDataB, 16);

			TeleToken.SetLastError(Err.ERROR_SUCCESS);
		}

		/*****************************************************************************
		Metoda zabezpiecza sektor w pamieci nieulotnej przed zapisem
		We:
			Adres na 16-bajtowy klucz do poczenia
			Obiekt poczenia
			Numer sektora do zabezpieczenia
		Wy:
			ustawiony "LastError"
		*****************************************************************************/
		static public void TVTTAPI_MemProtect(byte[] Pass, CTVTTAPI_INT_CON Con,
											  byte SNr)
		{
			uint Res, InKey = 0;
			byte[] OutDataA = new byte[16];
			byte[] OutDataB = new byte[16];
			byte[] InDataA = new byte[16];
			byte[] InDataB = new byte[16];

			if ((Pass == null) || (Con == null)) {
				TeleToken.SetLastError(Err.ERROR_INVALID_DATA);
				return;
			}

			TVTTAPI_Int_DecodeCon(Pass, Con);

			ZeroMem.ZeroMemory(ref OutDataA, 16);
			ZeroMem.ZeroMemory(ref OutDataB, 16);
			ZeroMem.ZeroMemory(ref InDataA, 16);
			ZeroMem.ZeroMemory(ref InDataB, 16);


			Con.OutKey(Con.OutKey()+1);
			OutDataB[TeleToken.TVTTD_CMD] = TeleToken.TVTTC_NVMEM_PROTECT;
			OutDataB[TeleToken.TVTTD_PARAM1] = SNr;

			TeleToken.TVTT_Calc(OutDataA, ref OutDataB);
			CTVTTAPI_Key Key = new CTVTTAPI_Key(Con);
			TeleToken.TVTT_CryptBlock(ref OutDataA, Con.OutKey(), Key.Data,
				TeleToken.TVTT_BLOCK_A, TeleToken.TVTT_PC2T);
			TeleToken.TVTT_CryptBlock(ref OutDataB, Con.OutKey(), Key.Data,
				TeleToken.TVTT_BLOCK_B, TeleToken.TVTT_PC2T);
			Res = TeleToken.TVTT_Talk(Con.hToken(), 5, Con.OutKey(),
				ref OutDataA[0], ref OutDataB[0], ref InKey,
				ref InDataA[0], ref InDataB[0]);
			TeleToken.TVTT_CryptBlock(ref InDataA, InKey, Key.Data,
				TeleToken.TVTT_BLOCK_A, TeleToken.TVTT_T2PC);
			TeleToken.TVTT_CryptBlock(ref InDataB, InKey, Key.Data,
				TeleToken.TVTT_BLOCK_B, TeleToken.TVTT_T2PC);
			Key.Free();
			if (TVTTAPI_Int_CheckError(Con, Res, InKey, InDataA, InDataB) != 0) {
				InKey = 0;
				ZeroMem.ZeroMemory(ref OutDataA, 16);
				ZeroMem.ZeroMemory(ref OutDataB, 16);
				ZeroMem.ZeroMemory(ref InDataA, 16);
				ZeroMem.ZeroMemory(ref InDataB, 16);
				TVTTAPI_Int_CodeCon(Pass, Con);
			}

			TVTTAPI_Int_CodeCon(Pass, Con);

			InKey = 0;
			ZeroMem.ZeroMemory(ref OutDataA, 16);
			ZeroMem.ZeroMemory(ref OutDataB, 16);
			ZeroMem.ZeroMemory(ref InDataA, 16);
			ZeroMem.ZeroMemory(ref InDataB, 16);

		  	TeleToken.SetLastError(Err.ERROR_SUCCESS);
		}

		/*****************************************************************************
		Metoda szyfruje blok danych o rozmiarze szesnastu bajtw przy uyciu Tokenu
		We:
			Adres na 16-bajtowy klucz do poczenia
			Obiekt poczenia
			Numer hasa do zaszyfrowania
			Tablica bufora z danymi do zaszyfrowania
		Wy:
			zaszyfrowane dane w buforze
			ustawiony "LastError"
		*****************************************************************************/
		static public void TVTTAPI_Encrypt(byte[] Pass, CTVTTAPI_INT_CON Con,
										   byte PNr, ref byte[] Buff)
		{
			uint Res, InKey = 0;
			byte[] OutDataA = new byte[16];
			byte[] OutDataB = new byte[16];
			byte[] InDataA = new byte[16];
			byte[] InDataB = new byte[16];

			if ((Pass == null) || (Con == null) || (Buff == null)) {
				TeleToken.SetLastError(Err.ERROR_INVALID_DATA);
				return;
			}


			TVTTAPI_Int_DecodeCon(Pass, Con);

			ZeroMem.CopyTable(ref OutDataA, Buff, 0, 0, 16);
			ZeroMem.ZeroMemory(ref OutDataB, 16);
			ZeroMem.ZeroMemory(ref InDataA, 16);
			ZeroMem.ZeroMemory(ref InDataB, 16);

			Con.OutKey(Con.OutKey()+1);
			OutDataB[TeleToken.TVTTD_CMD] = TeleToken.TVTTC_ENCRYPT;
			OutDataB[TeleToken.TVTTD_PARAM1] = PNr;

			TeleToken.TVTT_Calc(OutDataA, ref OutDataB);
			CTVTTAPI_Key Key = new CTVTTAPI_Key(Con);
			TeleToken.TVTT_CryptBlock(ref OutDataA, Con.OutKey(), Key.Data,
				TeleToken.TVTT_BLOCK_A, TeleToken.TVTT_PC2T);
			TeleToken.TVTT_CryptBlock(ref OutDataB, Con.OutKey(), Key.Data,
				TeleToken.TVTT_BLOCK_B, TeleToken.TVTT_PC2T);
			Res = TeleToken.TVTT_Talk(Con.hToken(), 5, Con.OutKey(),
				ref OutDataA[0], ref OutDataB[0], ref InKey,
				ref InDataA[0], ref InDataB[0]);
			TeleToken.TVTT_CryptBlock(ref InDataA, InKey, Key.Data,
				TeleToken.TVTT_BLOCK_A, TeleToken.TVTT_T2PC);
			TeleToken.TVTT_CryptBlock(ref InDataB, InKey, Key.Data,
				TeleToken.TVTT_BLOCK_B, TeleToken.TVTT_T2PC);
			Key.Free();
			if (TVTTAPI_Int_CheckError(Con, Res, InKey, InDataA, InDataB) != 0) {
				InKey = 0;
				ZeroMem.ZeroMemory(ref OutDataA, 16);
				ZeroMem.ZeroMemory(ref OutDataB, 16);
				ZeroMem.ZeroMemory(ref InDataA, 16);
				ZeroMem.ZeroMemory(ref InDataB, 16);
				TVTTAPI_Int_CodeCon(Pass, Con);
				return;
			}

			  ZeroMem.CopyTable(ref Buff, InDataA, 0, 0, 16);

			  TVTTAPI_Int_CodeCon(Pass, Con);

			  InKey = 0;
			  ZeroMem.ZeroMemory(ref OutDataA, 16);
			  ZeroMem.ZeroMemory(ref OutDataB, 16);
			  ZeroMem.ZeroMemory(ref InDataA, 16);
			  ZeroMem.ZeroMemory(ref InDataB, 16);

			  TeleToken.SetLastError(Err.ERROR_SUCCESS);
		}

		/*****************************************************************************
		Metoda deszyfruje blok szesnastu bajtw danych przy uyciu Tokenu
		We:
			Adres na 16-bajtowy klucz do poczenia
			Obiekt poczenia
			Numer hasa do zaszyfrowania
			Tablica bufora z danymi do zaszyfrowania
		Wy:
			zaszyfrowane dane w buforze
			ustawiony "LastError"
		*****************************************************************************/
		static public void TVTTAPI_Decrypt(byte[] Pass, CTVTTAPI_INT_CON Con,
										   byte PNr, ref byte[] Buff)
		{
			uint Res, InKey = 0;
			byte[] OutDataA = new byte[16];
			byte[] OutDataB = new byte[16];
			byte[] InDataA = new byte[16];
			byte[] InDataB = new byte[16];

			if ((Pass == null) || (Con == null) || (Buff == null)) {
				TeleToken.SetLastError(Err.ERROR_INVALID_DATA);
				return;
			}

			TVTTAPI_Int_DecodeCon(Pass, Con);

			ZeroMem.CopyTable(ref OutDataA, Buff, 0, 0, 16);
			ZeroMem.ZeroMemory(ref OutDataB, 16);
			ZeroMem.ZeroMemory(ref InDataA, 16);
			ZeroMem.ZeroMemory(ref InDataB, 16);

			Con.OutKey(Con.OutKey()+1);
			OutDataB[TeleToken.TVTTD_CMD] = TeleToken.TVTTC_DECRYPT;
			OutDataB[TeleToken.TVTTD_PARAM1] = PNr;

			TeleToken.TVTT_Calc(OutDataA, ref OutDataB);
			CTVTTAPI_Key Key = new CTVTTAPI_Key(Con);
			TeleToken.TVTT_CryptBlock(ref OutDataA, Con.OutKey(), Key.Data,
				TeleToken.TVTT_BLOCK_A, TeleToken.TVTT_PC2T);
			TeleToken.TVTT_CryptBlock(ref OutDataB, Con.OutKey(), Key.Data,
				TeleToken.TVTT_BLOCK_B, TeleToken.TVTT_PC2T);
			Res = TeleToken.TVTT_Talk(Con.hToken(), 5, Con.OutKey(),
				ref OutDataA[0], ref OutDataB[0], ref InKey,
				ref InDataA[0], ref InDataB[0]);
			TeleToken.TVTT_CryptBlock(ref InDataA, InKey, Key.Data,
				TeleToken.TVTT_BLOCK_A, TeleToken.TVTT_T2PC);
			TeleToken.TVTT_CryptBlock(ref InDataB, InKey, Key.Data,
				TeleToken.TVTT_BLOCK_B, TeleToken.TVTT_T2PC);
			Key.Free();
			if (TVTTAPI_Int_CheckError(Con, Res, InKey, InDataA, InDataB) != 0) {
				InKey = 0;
				ZeroMem.ZeroMemory(ref OutDataA, 16);
				ZeroMem.ZeroMemory(ref OutDataB, 16);
				ZeroMem.ZeroMemory(ref InDataA, 16);
				ZeroMem.ZeroMemory(ref InDataB, 16);
				TVTTAPI_Int_CodeCon(Pass, Con);
				return;
			}

			ZeroMem.CopyTable(ref Buff, InDataA, 0, 0, 16);

			TVTTAPI_Int_CodeCon(Pass, Con);

			InKey = 0;
			ZeroMem.ZeroMemory(ref OutDataA, 16);
			ZeroMem.ZeroMemory(ref OutDataB, 16);
			ZeroMem.ZeroMemory(ref InDataA, 16);
			ZeroMem.ZeroMemory(ref InDataB, 16);

		  	TeleToken.SetLastError(Err.ERROR_SUCCESS);
		}

		/*****************************************************************************
		Metoda odczytuje wartoc z licznika uniwersalnego
		We:
			Adres na 16-bajtowy klucz do poczenia
			Obiekt poczenia
			Numer licznika do odczytania
		Wy:
			Warto odczytana z licznika
			ustawiony "LastError"
		*****************************************************************************/
		static public uint TVTTAPI_CounterRead(byte[] Pass, CTVTTAPI_INT_CON Con,
										  byte CNr)
		{
			uint Res, InKey = 0;
			byte[] OutDataA = new byte[16];
			byte[] OutDataB = new byte[16];
			byte[] InDataA = new byte[16];
			byte[] InDataB = new byte[16];

			if ((Pass == null) || (Con == null)) {
				TeleToken.SetLastError(Err.ERROR_INVALID_DATA);
				return 0;
			}

			TVTTAPI_Int_DecodeCon(Pass, Con);

			ZeroMem.ZeroMemory(ref OutDataA, 16);
			ZeroMem.ZeroMemory(ref OutDataB, 16);
			ZeroMem.ZeroMemory(ref InDataA, 16);
			ZeroMem.ZeroMemory(ref InDataB, 16);

			Con.OutKey(Con.OutKey()+1);
			OutDataB[TeleToken.TVTTD_CMD] = TeleToken.TVTTC_CNT_READ;
			OutDataB[TeleToken.TVTTD_PARAM1] = CNr;

			TeleToken.TVTT_Calc(OutDataA, ref OutDataB);
			CTVTTAPI_Key Key = new CTVTTAPI_Key(Con);
			TeleToken.TVTT_CryptBlock(ref OutDataA, Con.OutKey(), Key.Data,
				TeleToken.TVTT_BLOCK_A, TeleToken.TVTT_PC2T);
			TeleToken.TVTT_CryptBlock(ref OutDataB, Con.OutKey(), Key.Data,
				TeleToken.TVTT_BLOCK_B, TeleToken.TVTT_PC2T);
			Res = TeleToken.TVTT_Talk(Con.hToken(), 5, Con.OutKey(),
				ref OutDataA[0], ref OutDataB[0], ref InKey,
				ref InDataA[0], ref InDataB[0]);
			TeleToken.TVTT_CryptBlock(ref InDataA, InKey, Key.Data,
				TeleToken.TVTT_BLOCK_A, TeleToken.TVTT_T2PC);
			TeleToken.TVTT_CryptBlock(ref InDataB, InKey, Key.Data,
				TeleToken.TVTT_BLOCK_B, TeleToken.TVTT_T2PC);
			Key.Free();
			if (TVTTAPI_Int_CheckError(Con, Res, InKey, InDataA, InDataB) != 0) {
				InKey = 0;
				ZeroMem.ZeroMemory(ref OutDataA, 16);
				ZeroMem.ZeroMemory(ref OutDataB, 16);
				ZeroMem.ZeroMemory(ref InDataA, 16);
				ZeroMem.ZeroMemory(ref InDataB, 16);
				TVTTAPI_Int_CodeCon(Pass, Con);
				return 0;
			}

			Res = InDataA[1];
			Res = Res << 8;
			Res |= InDataA[0];

			TVTTAPI_Int_CodeCon(Pass, Con);

			InKey = 0;
			ZeroMem.ZeroMemory(ref OutDataA, 16);
			ZeroMem.ZeroMemory(ref OutDataB, 16);
			ZeroMem.ZeroMemory(ref InDataA, 16);
			ZeroMem.ZeroMemory(ref InDataB, 16);

			TeleToken.SetLastError(Err.ERROR_SUCCESS);
			return Res;
		}

		/*****************************************************************************
		Metoda zwiksza licznik i odczytuje jego warto
		We:
			Adres na 16-bajtowy klucz do poczenia
			Obiekt poczenia
			Numer licznika do odczytania
		Wy:
			Warto odczytana z licznika
			ustawiony "LastError"
		*****************************************************************************/
		static public uint TVTTAPI_CounterIncRead(byte[] Pass,
												  CTVTTAPI_INT_CON Con,
												  byte CNr)
		{
			uint Res, InKey = 0;
			byte[] OutDataA = new byte[16];
			byte[] OutDataB = new byte[16];
			byte[] InDataA = new byte[16];
			byte[] InDataB = new byte[16];

			if ((Pass == null) || (Con == null)) {
				TeleToken.SetLastError(Err.ERROR_INVALID_DATA);
				return 0;
			}

			TVTTAPI_Int_DecodeCon(Pass, Con);

			ZeroMem.ZeroMemory(ref OutDataA, 16);
			ZeroMem.ZeroMemory(ref OutDataB, 16);
			ZeroMem.ZeroMemory(ref InDataA, 16);
			ZeroMem.ZeroMemory(ref InDataB, 16);

			Con.OutKey(Con.OutKey()+1);
			OutDataB[TeleToken.TVTTD_CMD] = TeleToken.TVTTC_CNT_INC_READ;
			OutDataB[TeleToken.TVTTD_PARAM1] = CNr;

			TeleToken.TVTT_Calc(OutDataA, ref OutDataB);
			CTVTTAPI_Key Key = new CTVTTAPI_Key(Con);
			TeleToken.TVTT_CryptBlock(ref OutDataA, Con.OutKey(), Key.Data,
				TeleToken.TVTT_BLOCK_A, TeleToken.TVTT_PC2T);
			TeleToken.TVTT_CryptBlock(ref OutDataB, Con.OutKey(), Key.Data,
				TeleToken.TVTT_BLOCK_B, TeleToken.TVTT_PC2T);
			Res = TeleToken.TVTT_Talk(Con.hToken(), 5, Con.OutKey(),
				ref OutDataA[0], ref OutDataB[0], ref InKey,
				ref InDataA[0], ref InDataB[0]);
			TeleToken.TVTT_CryptBlock(ref InDataA, InKey, Key.Data,
				TeleToken.TVTT_BLOCK_A, TeleToken.TVTT_T2PC);
			TeleToken.TVTT_CryptBlock(ref InDataB, InKey, Key.Data,
				TeleToken.TVTT_BLOCK_B, TeleToken.TVTT_T2PC);
			Key.Free();
			if (TVTTAPI_Int_CheckError(Con, Res, InKey, InDataA, InDataB) != 0) {
				InKey = 0;
				ZeroMem.ZeroMemory(ref OutDataA, 16);
				ZeroMem.ZeroMemory(ref OutDataB, 16);
				ZeroMem.ZeroMemory(ref InDataA, 16);
				ZeroMem.ZeroMemory(ref InDataB, 16);
				TVTTAPI_Int_CodeCon(Pass, Con);
				return 0;
			}

			Res = InDataA[1];
			Res = Res << 8;
			Res |= InDataA[0];

			TVTTAPI_Int_CodeCon(Pass, Con);

			InKey = 0;
			ZeroMem.ZeroMemory(ref OutDataA, 16);
			ZeroMem.ZeroMemory(ref OutDataB,16);
			ZeroMem.ZeroMemory(ref InDataA,16);
			ZeroMem.ZeroMemory(ref InDataB,16);

			TeleToken.SetLastError(Err.ERROR_SUCCESS);
			return Res;
		}

		/*****************************************************************************
		Metoda zmniejsza licznik i odczytuje jego warto
		We:
			Adres na 16-bajtowy klucz do poczenia
			Obiekt poczenia
			Numer licznika do odczytania
		Wy:
			Warto odczytana z licznika
			ustawiony "LastError"
		*****************************************************************************/
		static public uint TVTTAPI_CounterDecRead(byte[] Pass,
												  CTVTTAPI_INT_CON Con,
												  byte CNr)
		{
			uint Res, InKey = 0;
			byte[] OutDataA = new byte[16];
			byte[] OutDataB = new byte[16];
			byte[] InDataA = new byte[16];
			byte[] InDataB = new byte[16];

			if ((Pass == null) || (Con == null)) {
				TeleToken.SetLastError(Err.ERROR_INVALID_DATA);
				return 0;
			}

			TVTTAPI_Int_DecodeCon(Pass, Con);

			ZeroMem.ZeroMemory(ref OutDataA, 16);
			ZeroMem.ZeroMemory(ref OutDataB, 16);
			ZeroMem.ZeroMemory(ref InDataA, 16);
			ZeroMem.ZeroMemory(ref InDataB, 16);

			Con.OutKey(Con.OutKey()+1);
			OutDataB[TeleToken.TVTTD_CMD] = TeleToken.TVTTC_CNT_DEC_READ;
			OutDataB[TeleToken.TVTTD_PARAM1] = CNr;

			TeleToken.TVTT_Calc(OutDataA, ref OutDataB);
			CTVTTAPI_Key Key = new CTVTTAPI_Key(Con);
			TeleToken.TVTT_CryptBlock(ref OutDataA, Con.OutKey(), Key.Data,
				TeleToken.TVTT_BLOCK_A, TeleToken.TVTT_PC2T);
			TeleToken.TVTT_CryptBlock(ref OutDataB, Con.OutKey(), Key.Data,
				TeleToken.TVTT_BLOCK_B, TeleToken.TVTT_PC2T);
			Res = TeleToken.TVTT_Talk(Con.hToken(), 5, Con.OutKey(),
				ref OutDataA[0], ref OutDataB[0], ref InKey,
				ref InDataA[0], ref InDataB[0]);
			TeleToken.TVTT_CryptBlock(ref InDataA, InKey, Key.Data,
				TeleToken.TVTT_BLOCK_A, TeleToken.TVTT_T2PC);
			TeleToken.TVTT_CryptBlock(ref InDataB, InKey, Key.Data,
				TeleToken.TVTT_BLOCK_B, TeleToken.TVTT_T2PC);
			Key.Free();
			if (TVTTAPI_Int_CheckError(Con, Res, InKey, InDataA, InDataB) != 0) {
				InKey = 0;
				ZeroMem.ZeroMemory(ref OutDataA, 16);
				ZeroMem.ZeroMemory(ref OutDataB, 16);
				ZeroMem.ZeroMemory(ref InDataA, 16);
				ZeroMem.ZeroMemory(ref InDataB, 16);
				TVTTAPI_Int_CodeCon(Pass, Con);
				return 0;
			}

			Res = InDataA[1];
			Res = Res << 8;
			Res |= InDataA[0];

			TVTTAPI_Int_CodeCon(Pass, Con);

			InKey = 0;
			ZeroMem.ZeroMemory(ref OutDataA,16);
			ZeroMem.ZeroMemory(ref OutDataB,16);
			ZeroMem.ZeroMemory(ref InDataA,16);
			ZeroMem.ZeroMemory(ref InDataB,16);

			TeleToken.SetLastError(Err.ERROR_SUCCESS);
			return Res;
		}

		/*****************************************************************************
		Metoda zwiksza licznik, odczytuje jego zawartoc i odsya ja zaszyfrowan
		podanym hasem
		We:
			Adres na 16-bajtowy klucz do poczenia
			Obiekt poczenia
			Numer licznika do odczytania
			Numer hasa ktrym ma nastpi szyfrowanie
			Tablica bufora na dane
		Wy:
			wypeniony bufor na dane
			ustawiony "LastError"
		*****************************************************************************/
		static public void TVTTAPI_CounterIncEncryptRead(byte[] Pass,
														 CTVTTAPI_INT_CON Con,
														 byte CNr, byte PNr,
														 ref byte[] Buff)
		{
			uint Res, InKey = 0;
			byte[] OutDataA = new byte[16];
			byte[] OutDataB = new byte[16];
			byte[] InDataA = new byte[16];
			byte[] InDataB = new byte[16];

			if ((Pass == null) || (Con == null) || (Buff == null)) {
				TeleToken.SetLastError(Err.ERROR_INVALID_DATA);
				return;
			}

			ZeroMem.ZeroMemory(ref Buff, 16);

			TVTTAPI_Int_DecodeCon(Pass, Con);

			ZeroMem.ZeroMemory(ref OutDataA, 16);
			ZeroMem.ZeroMemory(ref OutDataB, 16);
			ZeroMem.ZeroMemory(ref InDataA, 16);
			ZeroMem.ZeroMemory(ref InDataB, 16);

			Con.OutKey(Con.OutKey()+1);
			OutDataB[TeleToken.TVTTD_CMD] = TeleToken.TVTTC_CNT_INC_ENCRYPT_READ;
			OutDataB[TeleToken.TVTTD_PARAM1] = CNr;
			OutDataB[TeleToken.TVTTD_PARAM2] = PNr;

			TeleToken.TVTT_Calc(OutDataA, ref OutDataB);
			CTVTTAPI_Key Key = new CTVTTAPI_Key(Con);
			TeleToken.TVTT_CryptBlock(ref OutDataA, Con.OutKey(), Key.Data,
				TeleToken.TVTT_BLOCK_A, TeleToken.TVTT_PC2T);
			TeleToken.TVTT_CryptBlock(ref OutDataB, Con.OutKey(), Key.Data,
				TeleToken.TVTT_BLOCK_B, TeleToken.TVTT_PC2T);
			Res = TeleToken.TVTT_Talk(Con.hToken(), 5, Con.OutKey(),
				ref OutDataA[0], ref OutDataB[0], ref InKey,
				ref InDataA[0], ref InDataB[0]);
			TeleToken.TVTT_CryptBlock(ref InDataA, InKey, Key.Data,
				TeleToken.TVTT_BLOCK_A, TeleToken.TVTT_T2PC);
			TeleToken.TVTT_CryptBlock(ref InDataB, InKey, Key.Data,
				TeleToken.TVTT_BLOCK_B, TeleToken.TVTT_T2PC);
			Key.Free();
			if (TVTTAPI_Int_CheckError(Con, Res, InKey, InDataA, InDataB) != 0) {
				InKey = 0;
				ZeroMem.ZeroMemory(ref OutDataA, 16);
				ZeroMem.ZeroMemory(ref OutDataB, 16);
				ZeroMem.ZeroMemory(ref InDataA, 16);
				ZeroMem.ZeroMemory(ref InDataB, 16);
				TVTTAPI_Int_CodeCon(Pass, Con);
			}

			ZeroMem.CopyTable(ref Buff, InDataA, 0, 0, 16);

			TVTTAPI_Int_CodeCon(Pass, Con);

			InKey = 0;
			ZeroMem.ZeroMemory(ref OutDataA, 16);
			ZeroMem.ZeroMemory(ref OutDataB, 16);
			ZeroMem.ZeroMemory(ref InDataA, 16);
			ZeroMem.ZeroMemory(ref InDataB, 16);

			TeleToken.SetLastError(Err.ERROR_SUCCESS);
		}

		/*****************************************************************************
		Metoda zapisuje warto do licznika uniwersalnego
		We:
			Adres na 16-bajtowy klucz do poczenia
			Obiekt poczenia
			Numer licznika do zapisania
			Wartoc do zapisania
		Wy:
			ustawiony "LastError"
		*****************************************************************************/
		static public void TVTTAPI_CounterWrite(byte[] Pass,
												CTVTTAPI_INT_CON Con,
												byte CNr, uint Value)
		{
			uint Res, InKey = 0;
			byte[] OutDataA = new byte[16];
			byte[] OutDataB = new byte[16];
			byte[] InDataA = new byte[16];
			byte[] InDataB = new byte[16];

			if ((Pass == null) || (Con == null)) {
				TeleToken.SetLastError(Err.ERROR_INVALID_DATA);
				return;
			}

			TVTTAPI_Int_DecodeCon(Pass, Con);

			ZeroMem.ZeroMemory(ref OutDataA, 16);
			ZeroMem.ZeroMemory(ref OutDataB, 16);
			ZeroMem.ZeroMemory(ref InDataA, 16);
			ZeroMem.ZeroMemory(ref InDataB, 16);

			OutDataA[0] = (byte)(Value & 0xff);
			Value = Value >> 8;
			OutDataA[1] = (byte)(Value & 0xff);
			Value = 0;

			Con.OutKey(Con.OutKey()+1);
			OutDataB[TeleToken.TVTTD_CMD] = TeleToken.TVTTC_CNT_WRITE;
			OutDataB[TeleToken.TVTTD_PARAM1] = CNr;

			TeleToken.TVTT_Calc(OutDataA, ref OutDataB);
			CTVTTAPI_Key Key = new CTVTTAPI_Key(Con);
			TeleToken.TVTT_CryptBlock(ref OutDataA, Con.OutKey(), Key.Data,
				TeleToken.TVTT_BLOCK_A, TeleToken.TVTT_PC2T);
			TeleToken.TVTT_CryptBlock(ref OutDataB, Con.OutKey(), Key.Data,
				TeleToken.TVTT_BLOCK_B, TeleToken.TVTT_PC2T);
			Res = TeleToken.TVTT_Talk(Con.hToken(), 5, Con.OutKey(),
				ref OutDataA[0], ref OutDataB[0], ref InKey,
				ref InDataA[0], ref InDataB[0]);
			TeleToken.TVTT_CryptBlock(ref InDataA, InKey, Key.Data,
				TeleToken.TVTT_BLOCK_A, TeleToken.TVTT_T2PC);
			TeleToken.TVTT_CryptBlock(ref InDataB, InKey, Key.Data,
				TeleToken.TVTT_BLOCK_B, TeleToken.TVTT_T2PC);
			Key.Free();
			if (TVTTAPI_Int_CheckError(Con, Res, InKey, InDataA, InDataB) != 0) {
				InKey = 0;
				ZeroMem.ZeroMemory(ref OutDataA, 16);
				ZeroMem.ZeroMemory(ref OutDataB, 16);
				ZeroMem.ZeroMemory(ref InDataA, 16);
				ZeroMem.ZeroMemory(ref InDataB, 16);
				TVTTAPI_Int_CodeCon(Pass, Con);
				return;
			}

			TVTTAPI_Int_CodeCon(Pass, Con);

			InKey = 0;
			ZeroMem.ZeroMemory(ref OutDataA, 16);
			ZeroMem.ZeroMemory(ref OutDataB, 16);
			ZeroMem.ZeroMemory(ref InDataA, 16);
			ZeroMem.ZeroMemory(ref InDataB, 16);

		  	TeleToken.SetLastError(Err.ERROR_SUCCESS);
		}

		/*****************************************************************************
		Metoda zapisuje dwa bajty konfiguracji licznika
		We:
			Adres na 16-bajtowy klucz do poczenia
			Obiekt poczenia
			Numer licznika do zapisania konfiguracji
			Wartoc pierwszego bajtu konfiguracji
			Wartoc drugiego bajtu konfiguracji
		Wy:
			ustawiony "LastError"
		*****************************************************************************/
		static public void TVTTAPI_CounterConfigWrite(byte[] Pass,
													  CTVTTAPI_INT_CON Con,
													  byte CNr, byte Config0,
													  byte Config1)
		{
			uint Res, InKey = 0;
			byte[] OutDataA = new byte[16];
			byte[] OutDataB = new byte[16];
			byte[] InDataA = new byte[16];
			byte[] InDataB = new byte[16];

			if ((Pass == null) || (Con == null)) {
				TeleToken.SetLastError(Err.ERROR_INVALID_DATA);
				return;
			}

			TVTTAPI_Int_DecodeCon(Pass, Con);

			ZeroMem.ZeroMemory(ref OutDataA, 16);
			ZeroMem.ZeroMemory(ref OutDataB, 16);
			ZeroMem.ZeroMemory(ref InDataA, 16);
			ZeroMem.ZeroMemory(ref InDataB, 16);

			OutDataA[0] = Config0;
			OutDataA[1] = Config1;

			Con.OutKey(Con.OutKey()+1);
			OutDataB[TeleToken.TVTTD_CMD] = TeleToken.TVTTC_CNT_WRITE_CONFIG;
			OutDataB[TeleToken.TVTTD_PARAM1] = CNr;

			TeleToken.TVTT_Calc(OutDataA, ref OutDataB);
			CTVTTAPI_Key Key = new CTVTTAPI_Key(Con);
			TeleToken.TVTT_CryptBlock(ref OutDataA, Con.OutKey(), Key.Data,
				TeleToken.TVTT_BLOCK_A, TeleToken.TVTT_PC2T);
			TeleToken.TVTT_CryptBlock(ref OutDataB, Con.OutKey(), Key.Data,
				TeleToken.TVTT_BLOCK_B, TeleToken.TVTT_PC2T);
			Res = TeleToken.TVTT_Talk(Con.hToken(), 5, Con.OutKey(),
				ref OutDataA[0], ref OutDataB[0], ref InKey,
				ref InDataA[0], ref InDataB[0]);
			TeleToken.TVTT_CryptBlock(ref InDataA, InKey, Key.Data,
				TeleToken.TVTT_BLOCK_A, TeleToken.TVTT_T2PC);
			TeleToken.TVTT_CryptBlock(ref InDataB, InKey, Key.Data,
				TeleToken.TVTT_BLOCK_B, TeleToken.TVTT_T2PC);
			Key.Free();
			if (TVTTAPI_Int_CheckError(Con, Res, InKey, InDataA, InDataB) != 0) {
				InKey = 0;
				ZeroMem.ZeroMemory(ref OutDataA, 16);
				ZeroMem.ZeroMemory(ref OutDataB, 16);
				ZeroMem.ZeroMemory(ref InDataA, 16);
				ZeroMem.ZeroMemory(ref InDataB, 16);
				TVTTAPI_Int_CodeCon(Pass, Con);
				return;
			}

			TVTTAPI_Int_CodeCon(Pass, Con);

			InKey = 0;
			ZeroMem.ZeroMemory(ref OutDataA, 16);
			ZeroMem.ZeroMemory(ref OutDataB, 16);
			ZeroMem.ZeroMemory(ref InDataA, 16);
			ZeroMem.ZeroMemory(ref InDataB, 16);

		  	TeleToken.SetLastError(Err.ERROR_SUCCESS);
		}

		/*****************************************************************************
		Metoda ustawia bity w dwch bajtach konfiguracji licznika
		We:
			Adres na 16-bajtowy klucz do poczenia
			Obiekt poczenia
			Numer licznika do zapisania konfiguracji
			Maska bitowa pierwszego bajtu konfiguracji
			Maska bitowa pierwszego bajtu konfiguracji
		Wy:
			ustawiony "LastError"
		*****************************************************************************/
		static public void TVTTAPI_CounterConfigSet(byte[] Pass,
													CTVTTAPI_INT_CON Con,
													byte CNr, byte Config0,
													byte Config1)
		{
			uint Res, InKey = 0;
			byte[] OutDataA = new byte[16];
			byte[] OutDataB = new byte[16];
			byte[] InDataA = new byte[16];
			byte[] InDataB = new byte[16];

			if ((Pass == null) || (Con == null)) {
				TeleToken.SetLastError(Err.ERROR_INVALID_DATA);
				return;
			}

			TVTTAPI_Int_DecodeCon(Pass, Con);

			ZeroMem.ZeroMemory(ref OutDataA, 16);
			ZeroMem.ZeroMemory(ref OutDataB, 16);
			ZeroMem.ZeroMemory(ref InDataA, 16);
			ZeroMem.ZeroMemory(ref InDataB, 16);

			OutDataA[0] = Config0;
			OutDataA[1] = Config1;

			Con.OutKey(Con.OutKey()+1);
			OutDataB[TeleToken.TVTTD_CMD] = TeleToken.TVTTC_CNT_SET_CONFIG;
			OutDataB[TeleToken.TVTTD_PARAM1] = CNr;

			TeleToken.TVTT_Calc(OutDataA, ref OutDataB);
			CTVTTAPI_Key Key = new CTVTTAPI_Key(Con);
			TeleToken.TVTT_CryptBlock(ref OutDataA, Con.OutKey(), Key.Data,
				TeleToken.TVTT_BLOCK_A, TeleToken.TVTT_PC2T);
			TeleToken.TVTT_CryptBlock(ref OutDataB, Con.OutKey(), Key.Data,
				TeleToken.TVTT_BLOCK_B, TeleToken.TVTT_PC2T);
			Res = TeleToken.TVTT_Talk(Con.hToken(), 5, Con.OutKey(),
				ref OutDataA[0], ref OutDataB[0], ref InKey,
				ref InDataA[0], ref InDataB[0]);
			TeleToken.TVTT_CryptBlock(ref InDataA, InKey, Key.Data,
				TeleToken.TVTT_BLOCK_A, TeleToken.TVTT_T2PC);
			TeleToken.TVTT_CryptBlock(ref InDataB, InKey, Key.Data,
				TeleToken.TVTT_BLOCK_B, TeleToken.TVTT_T2PC);
			Key.Free();
			if (TVTTAPI_Int_CheckError(Con, Res, InKey, InDataA, InDataB) != 0) {
				InKey = 0;
				ZeroMem.ZeroMemory(ref OutDataA, 16);
				ZeroMem.ZeroMemory(ref OutDataB, 16);
				ZeroMem.ZeroMemory(ref InDataA, 16);
				ZeroMem.ZeroMemory(ref InDataB, 16);
				TVTTAPI_Int_CodeCon(Pass, Con);
				return;
			}

			TVTTAPI_Int_CodeCon(Pass, Con);

			InKey = 0;
			ZeroMem.ZeroMemory(ref OutDataA, 16);
			ZeroMem.ZeroMemory(ref OutDataB, 16);
			ZeroMem.ZeroMemory(ref InDataA, 16);
			ZeroMem.ZeroMemory(ref InDataB, 16);

			TeleToken.SetLastError(Err.ERROR_SUCCESS);
		}

		/*****************************************************************************
		Metoda kasuje bity w dwch bajtach konfiguracji licznika
		We:
			Adres na 16-bajtowy klucz do poczenia
			Obiekt poczenia
			Numer licznika do zapisania konfiguracji
			Maska bitowa pierwszego bajtu konfiguracji
			Maska bitowa pierwszego bajtu konfiguracji
		Wy:
			ustawiony "LastError"
		*****************************************************************************/
		static public void TVTTAPI_CounterConfigClear(byte[] Pass,
													  CTVTTAPI_INT_CON Con,
													  byte CNr, byte Config0,
													  byte Config1)
		{
			uint Res, InKey = 0;
			byte[] OutDataA = new byte[16];
			byte[] OutDataB = new byte[16];
			byte[] InDataA = new byte[16];
			byte[] InDataB = new byte[16];

			if ((Pass == null) || (Con == null)) {
				TeleToken.SetLastError(Err.ERROR_INVALID_DATA);
				return;
			}

			TVTTAPI_Int_DecodeCon(Pass, Con);

			ZeroMem.ZeroMemory(ref OutDataA, 16);
			ZeroMem.ZeroMemory(ref OutDataB, 16);
			ZeroMem.ZeroMemory(ref InDataA, 16);
			ZeroMem.ZeroMemory(ref InDataB, 16);

			OutDataA[0] = Config0;
		  	OutDataA[1] = Config1;

			Con.OutKey(Con.OutKey()+1);
			OutDataB[TeleToken.TVTTD_CMD] = TeleToken.TVTTC_CNT_CLR_CONFIG;
			OutDataB[TeleToken.TVTTD_PARAM1] = CNr;

			TeleToken.TVTT_Calc(OutDataA, ref OutDataB);
			CTVTTAPI_Key Key = new CTVTTAPI_Key(Con);
			TeleToken.TVTT_CryptBlock(ref OutDataA, Con.OutKey(), Key.Data,
				TeleToken.TVTT_BLOCK_A, TeleToken.TVTT_PC2T);
			TeleToken.TVTT_CryptBlock(ref OutDataB, Con.OutKey(), Key.Data,
				TeleToken.TVTT_BLOCK_B, TeleToken.TVTT_PC2T);
			Res = TeleToken.TVTT_Talk(Con.hToken(), 5, Con.OutKey(),
				ref OutDataA[0], ref OutDataB[0], ref InKey,
				ref InDataA[0], ref InDataB[0]);
			TeleToken.TVTT_CryptBlock(ref InDataA, InKey, Key.Data,
				TeleToken.TVTT_BLOCK_A, TeleToken.TVTT_T2PC);
			TeleToken.TVTT_CryptBlock(ref InDataB, InKey, Key.Data,
				TeleToken.TVTT_BLOCK_B, TeleToken.TVTT_T2PC);
			Key.Free();
			if (TVTTAPI_Int_CheckError(Con, Res, InKey, InDataA, InDataB) != 0) {
				InKey = 0;
				ZeroMem.ZeroMemory(ref OutDataA, 16);
				ZeroMem.ZeroMemory(ref OutDataB, 16);
				ZeroMem.ZeroMemory(ref InDataA, 16);
				ZeroMem.ZeroMemory(ref InDataB, 16);
				TVTTAPI_Int_CodeCon(Pass, Con);
				return;
			}

			TVTTAPI_Int_CodeCon(Pass, Con);

			InKey = 0;
			ZeroMem.ZeroMemory(ref OutDataA, 16);
			ZeroMem.ZeroMemory(ref OutDataB, 16);
			ZeroMem.ZeroMemory(ref InDataA, 16);
			ZeroMem.ZeroMemory(ref InDataB, 16);

		  	TeleToken.SetLastError(Err.ERROR_SUCCESS);
		}

		/*****************************************************************************
		Metoda zapisuje haso do tablicy hase
		We:
			Adres na 16-bajtowy klucz do poczenia
			Obiekt poczenia
			Numer pozycji w tablicy hase
			Tablica bufora z hasem do zapisania
		Wy:
			ustawiony "LastError"
		*****************************************************************************/
		static public void TVTTAPI_PassWrite(byte[] Pass, CTVTTAPI_INT_CON Con,
											 byte PNr, byte[] Buff)
		{
			uint Res, InKey = 0;
			byte[] OutDataA = new byte[16];
			byte[] OutDataB = new byte[16];
			byte[] InDataA = new byte[16];
			byte[] InDataB = new byte[16];

			if ((Pass == null) || (Con == null) || (Buff == null)) {
				TeleToken.SetLastError(Err.ERROR_INVALID_DATA);
				return;
			}

			TVTTAPI_Int_DecodeCon(Pass, Con);

			ZeroMem.CopyTable(ref OutDataA, Buff, 0, 0, 16);
			ZeroMem.ZeroMemory(ref OutDataB, 16);
			ZeroMem.ZeroMemory(ref InDataA, 16);
			ZeroMem.ZeroMemory(ref InDataB, 16);

			Con.OutKey(Con.OutKey()+1);
			OutDataB[TeleToken.TVTTD_CMD] = TeleToken.TVTTC_PASS_WRITE;
			OutDataB[TeleToken.TVTTD_PARAM1] = PNr;

			TeleToken.TVTT_Calc(OutDataA, ref OutDataB);
			CTVTTAPI_Key Key = new CTVTTAPI_Key(Con);
			TeleToken.TVTT_CryptBlock(ref OutDataA, Con.OutKey(), Key.Data,
				TeleToken.TVTT_BLOCK_A, TeleToken.TVTT_PC2T);
			TeleToken.TVTT_CryptBlock(ref OutDataB, Con.OutKey(), Key.Data,
				TeleToken.TVTT_BLOCK_B, TeleToken.TVTT_PC2T);
			Res = TeleToken.TVTT_Talk(Con.hToken(), 5, Con.OutKey(),
				ref OutDataA[0], ref OutDataB[0], ref InKey,
				ref InDataA[0], ref InDataB[0]);
			TeleToken.TVTT_CryptBlock(ref InDataA, InKey, Key.Data,
				TeleToken.TVTT_BLOCK_A, TeleToken.TVTT_T2PC);
			TeleToken.TVTT_CryptBlock(ref InDataB, InKey, Key.Data,
				TeleToken.TVTT_BLOCK_B, TeleToken.TVTT_T2PC);
			Key.Free();
			if (TVTTAPI_Int_CheckError(Con, Res, InKey, InDataA, InDataB) != 0) {
				InKey = 0;
				ZeroMem.ZeroMemory(ref OutDataA, 16);
				ZeroMem.ZeroMemory(ref OutDataB, 16);
				ZeroMem.ZeroMemory(ref InDataA, 16);
				ZeroMem.ZeroMemory(ref InDataB, 16);
				TVTTAPI_Int_CodeCon(Pass, Con);
				return;
			}

			TVTTAPI_Int_CodeCon(Pass, Con);

			InKey = 0;
			ZeroMem.ZeroMemory(ref OutDataA, 16);
			ZeroMem.ZeroMemory(ref OutDataB, 16);
			ZeroMem.ZeroMemory(ref InDataA, 16);
			ZeroMem.ZeroMemory(ref InDataB, 16);

		  	TeleToken.SetLastError(Err.ERROR_SUCCESS);
		}

		/*****************************************************************************
		Metoda zapisuje dwa bajty konfiguracji hasa
		We:
			Adres na 16-bajtowy klucz do poczenia
			Obiekt poczenia
			Numer hasa do zapisania konfiguracji
			Wartoc pierwszego bajtu konfiguracji
			Wartoc drugiego bajtu konfiguracji
		Wy:
			ustawiony "LastError"
		*****************************************************************************/
		static public void TVTTAPI_PassConfigWrite(byte[] Pass,
												   CTVTTAPI_INT_CON Con,
												   byte PNr, byte Config0,
												   byte Config1)
		{
			uint Res, InKey = 0;
			byte[] OutDataA = new byte[16];
			byte[] OutDataB = new byte[16];
			byte[] InDataA = new byte[16];
			byte[] InDataB = new byte[16];

			if ((Pass == null) || (Con == null)) {
				TeleToken.SetLastError(Err.ERROR_INVALID_DATA);
				return;
			}

			TVTTAPI_Int_DecodeCon(Pass, Con);

			ZeroMem.ZeroMemory(ref OutDataA, 16);
			ZeroMem.ZeroMemory(ref OutDataB, 16);
			ZeroMem.ZeroMemory(ref InDataA, 16);
			ZeroMem.ZeroMemory(ref InDataB, 16);

			OutDataA[0] = Config0;
			OutDataA[1] = Config1;

			Con.OutKey(Con.OutKey()+1);
			OutDataB[TeleToken.TVTTD_CMD] = TeleToken.TVTTC_PASS_WRITE_CONFIG;
			OutDataB[TeleToken.TVTTD_PARAM1] = PNr;

			TeleToken.TVTT_Calc(OutDataA, ref OutDataB);
			CTVTTAPI_Key Key = new CTVTTAPI_Key(Con);
			TeleToken.TVTT_CryptBlock(ref OutDataA, Con.OutKey(), Key.Data,
				TeleToken.TVTT_BLOCK_A, TeleToken.TVTT_PC2T);
			TeleToken.TVTT_CryptBlock(ref OutDataB, Con.OutKey(), Key.Data,
				TeleToken.TVTT_BLOCK_B, TeleToken.TVTT_PC2T);
			Res = TeleToken.TVTT_Talk(Con.hToken(), 5, Con.OutKey(),
				ref OutDataA[0], ref OutDataB[0], ref InKey,
				ref InDataA[0], ref InDataB[0]);
			TeleToken.TVTT_CryptBlock(ref InDataA, InKey, Key.Data,
				TeleToken.TVTT_BLOCK_A, TeleToken.TVTT_T2PC);
			TeleToken.TVTT_CryptBlock(ref InDataB, InKey, Key.Data,
				TeleToken.TVTT_BLOCK_B, TeleToken.TVTT_T2PC);
			Key.Free();
			if (TVTTAPI_Int_CheckError(Con, Res, InKey, InDataA, InDataB) != 0) {
				InKey = 0;
				ZeroMem.ZeroMemory(ref OutDataA, 16);
				ZeroMem.ZeroMemory(ref OutDataB, 16);
				ZeroMem.ZeroMemory(ref InDataA, 16);
				ZeroMem.ZeroMemory(ref InDataB, 16);
				TVTTAPI_Int_CodeCon(Pass, Con);
				return;
			}

			TVTTAPI_Int_CodeCon(Pass, Con);

			InKey = 0;
			ZeroMem.ZeroMemory(ref OutDataA, 16);
			ZeroMem.ZeroMemory(ref OutDataB, 16);
			ZeroMem.ZeroMemory(ref InDataA, 16);
			ZeroMem.ZeroMemory(ref InDataB, 16);

			TeleToken.SetLastError(Err.ERROR_SUCCESS);
		}

		/*****************************************************************************
		Metoda ustawia bity w dwuch bajtach konfiguracji hasa
		We:
			Adres na 16-bajtowy klucz do poczenia
			Obiekt poczenia
			Numer hasa do zapisania konfiguracji
			Maska bitowa pierwszego bajtu konfiguracji
			Maska bitowa pierwszego bajtu konfiguracji
		Wy:
			ustawiony "LastError"
		*****************************************************************************/
		static public void TVTTAPI_PassConfigSet(byte[] Pass,
												 CTVTTAPI_INT_CON Con,
												 byte PNr, byte Config0,
												 byte Config1)
		{
			uint Res, InKey = 0;
			byte[] OutDataA = new byte[16];
			byte[] OutDataB = new byte[16];
			byte[] InDataA = new byte[16];
			byte[] InDataB = new byte[16];

			if ((Pass == null) || (Con == null)) {
				TeleToken.SetLastError(Err.ERROR_INVALID_DATA);
				return;
			}

			TVTTAPI_Int_DecodeCon(Pass, Con);

			ZeroMem.ZeroMemory(ref OutDataA, 16);
			ZeroMem.ZeroMemory(ref OutDataB, 16);
			ZeroMem.ZeroMemory(ref InDataA, 16);
			ZeroMem.ZeroMemory(ref InDataB, 16);

			OutDataA[0] = Config0;
			OutDataA[1] = Config1;

			Con.OutKey(Con.OutKey()+1);
			OutDataB[TeleToken.TVTTD_CMD] = TeleToken.TVTTC_PASS_SET_CONFIG;
			OutDataB[TeleToken.TVTTD_PARAM1] = PNr;

			TeleToken.TVTT_Calc(OutDataA, ref OutDataB);
			CTVTTAPI_Key Key = new CTVTTAPI_Key(Con);
			TeleToken.TVTT_CryptBlock(ref OutDataA, Con.OutKey(), Key.Data,
				TeleToken.TVTT_BLOCK_A, TeleToken.TVTT_PC2T);
			TeleToken.TVTT_CryptBlock(ref OutDataB, Con.OutKey(), Key.Data,
				TeleToken.TVTT_BLOCK_B, TeleToken.TVTT_PC2T);
			Res = TeleToken.TVTT_Talk(Con.hToken(), 5, Con.OutKey(),
				ref OutDataA[0], ref OutDataB[0], ref InKey,
				ref InDataA[0], ref InDataB[0]);
			TeleToken.TVTT_CryptBlock(ref InDataA, InKey, Key.Data,
				TeleToken.TVTT_BLOCK_A, TeleToken.TVTT_T2PC);
			TeleToken.TVTT_CryptBlock(ref InDataB, InKey, Key.Data,
				TeleToken.TVTT_BLOCK_B, TeleToken.TVTT_T2PC);
			Key.Free();
			if (TVTTAPI_Int_CheckError(Con, Res, InKey, InDataA, InDataB) != 0) {
				InKey = 0;
				ZeroMem.ZeroMemory(ref OutDataA, 16);
				ZeroMem.ZeroMemory(ref OutDataB, 16);
				ZeroMem.ZeroMemory(ref InDataA, 16);
				ZeroMem.ZeroMemory(ref InDataB, 16);
				TVTTAPI_Int_CodeCon(Pass, Con);
				return;
			}

			TVTTAPI_Int_CodeCon(Pass, Con);

			InKey = 0;
			ZeroMem.ZeroMemory(ref OutDataA, 16);
			ZeroMem.ZeroMemory(ref OutDataB, 16);
			ZeroMem.ZeroMemory(ref InDataA, 16);
			ZeroMem.ZeroMemory(ref InDataB, 16);

			TeleToken.SetLastError(Err.ERROR_SUCCESS);
		}

		/*****************************************************************************
		Metoda kasuje bity w dwuch bajtach konfiguracji hasa
		We:
			Adres na 16-bajtowy klucz do poczenia
			Obiekt poczenia
			Numer hasa do zapisania konfiguracji
			Maska bitowa pierwszego bajtu konfiguracji
			Maska bitowa pierwszego bajtu konfiguracji
		Wy:
			ustawiony "LastError"
		*****************************************************************************/
		static public void TVTTAPI_PassConfigClear(byte[] Pass,
												   CTVTTAPI_INT_CON Con,
												   byte PNr, byte Config0,
												   byte Config1)
		{
			uint Res, InKey = 0;
			byte[] OutDataA = new byte[16];
			byte[] OutDataB = new byte[16];
			byte[] InDataA = new byte[16];
			byte[] InDataB = new byte[16];

			if ((Pass == null) || (Con == null)) {
				TeleToken.SetLastError(Err.ERROR_INVALID_DATA);
				return;
			}

			TVTTAPI_Int_DecodeCon(Pass, Con);

			ZeroMem.ZeroMemory(ref OutDataA, 16);
			ZeroMem.ZeroMemory(ref OutDataB, 16);
			ZeroMem.ZeroMemory(ref InDataA, 16);
			ZeroMem.ZeroMemory(ref InDataB, 16);

			OutDataA[0] = Config0;
			OutDataA[1] = Config1;

		  	Con.OutKey(Con.OutKey()+1);
			OutDataB[TeleToken.TVTTD_CMD] = TeleToken.TVTTC_PASS_CLR_CONFIG;
			OutDataB[TeleToken.TVTTD_PARAM1] = PNr;

			TeleToken.TVTT_Calc(OutDataA, ref OutDataB);
			CTVTTAPI_Key Key = new CTVTTAPI_Key(Con);
			TeleToken.TVTT_CryptBlock(ref OutDataA, Con.OutKey(), Key.Data,
				TeleToken.TVTT_BLOCK_A, TeleToken.TVTT_PC2T);
			TeleToken.TVTT_CryptBlock(ref OutDataB, Con.OutKey(), Key.Data,
				TeleToken.TVTT_BLOCK_B, TeleToken.TVTT_PC2T);
			Res = TeleToken.TVTT_Talk(Con.hToken(), 5, Con.OutKey(),
				ref OutDataA[0], ref OutDataB[0], ref InKey,
				ref InDataA[0], ref InDataB[0]);
			TeleToken.TVTT_CryptBlock(ref InDataA, InKey, Key.Data,
				TeleToken.TVTT_BLOCK_A, TeleToken.TVTT_T2PC);
			TeleToken.TVTT_CryptBlock(ref InDataB, InKey, Key.Data,
				TeleToken.TVTT_BLOCK_B, TeleToken.TVTT_T2PC);
			Key.Free();
			if (TVTTAPI_Int_CheckError(Con, Res, InKey, InDataA, InDataB) != 0) {
				InKey = 0;
				ZeroMem.ZeroMemory(ref OutDataA, 16);
				ZeroMem.ZeroMemory(ref OutDataB, 16);
				ZeroMem.ZeroMemory(ref InDataA, 16);
				ZeroMem.ZeroMemory(ref InDataB, 16);
				TVTTAPI_Int_CodeCon(Pass, Con);
			}

			TVTTAPI_Int_CodeCon(Pass, Con);

			InKey = 0;
			ZeroMem.ZeroMemory(ref OutDataA, 16);
			ZeroMem.ZeroMemory(ref OutDataB, 16);
			ZeroMem.ZeroMemory(ref InDataA, 16);
			ZeroMem.ZeroMemory(ref InDataB, 16);

			TeleToken.SetLastError(Err.ERROR_SUCCESS);
		}

		/*****************************************************************************
		Metoda oblicza wielko tablicy potrzebnej na dane po zaszyfrowaniu
		We:
			Rozmiar tablicy do zaszyfrowania
		Wy:
			Rozmiar tablicy po zaszyfrowaniu
		*****************************************************************************/
		static public uint TVTTAPI_EncryptBlockSize(uint Size)
		{
			uint Res = Size / 16;
			if ((Size % 16) > 0) Res++;
			Res += 4;
			Res *= 16;
			return Res;
		}

		/*****************************************************************************
		Metoda szyfruje blok danych o dowolnym rozmiarze
		We:
			Adres na 16-bajtowy klucz do poczenia
			Obiekt poczenia
			Numer hasa do szyfrowania
			Tablica z danymi
			Ilo danych do zaszyfrowania
			Adres na zmienn do ktrej zostanie wpisana ilo danych po zaszyfrowaniu
		Wy:
			True jesli operacja sie powioda, w przeciwnym wypadku False
			Rozmiar bloku zaszyfrowanego (przez var OSize)
			ustawiony "LastError"
		*****************************************************************************/
		static public bool TVTTAPI_EncryptBlock(byte[] Pass,
												CTVTTAPI_INT_CON Con,
												byte PNr,
												ref byte[] Buff,
												uint Size,
												ref uint OSize)
		{
			uint n, i;
			byte[] Key = new byte[16];

			if ((Pass == null) || (Con == null) || (Buff == null) ||
				(Size <= 64)) {
				TeleToken.SetLastError(Err.ERROR_INVALID_DATA);
				return false;
			}

			OSize = TVTTAPI_EncryptBlockSize(Size);

			ZeroMem.CopyTable(ref Buff, Buff, 64, 0, (int) Size);
			TeleToken.TVTT_Rand(ref Buff);
			TeleToken.TVTT_Rand(ref Buff, 2*16);
			TeleToken.TVTT_Rand(ref Buff, 3*16);

			Buff[16+0] = (byte)(Size & 0xff);
			Buff[16+1] = (byte)((Size >> 8) & 0xff);
			Buff[16+2] = (byte)((Size >> 16) & 0xff);
			Buff[16+3] = (byte)((Size >> 24) & 0xff);

			Buff[16+4] = (byte)(OSize & 0xff);
			Buff[16+5] = (byte)((OSize >> 8) & 0xff);
			Buff[16+6] = (byte)((OSize >> 16) & 0xff);
			Buff[16+7] = (byte)((OSize >> 24) & 0xff);

			ZeroMem.CopyTable(ref Key, Buff, 0, 0, 16);
			TVTTAPI_Encrypt(Pass, Con, PNr, ref Key);
			if (TeleToken.GetLastError() != Err.ERROR_SUCCESS) {
				ZeroMem.ZeroMemory(ref Buff, (int) OSize);
				OSize = 0;
				return false;
			}

			rijndael.rijndaelEncrypt(Key, ref Buff, 16);

			ZeroMem.CopyTable(ref Key, Buff, 0, 2*16, 16);
			TVTTAPI_Encrypt(Pass, Con, PNr, ref Key);
			if (TeleToken.GetLastError() != Err.ERROR_SUCCESS) {
				ZeroMem.ZeroMemory(ref Buff, (int) OSize);
				OSize = 0;
				return false;
			}

			n = 4*16;
			while (n<OSize) {
				i = 0;
				while (i<16) {
				  Buff[n+i] ^= Buff[n+i-16];
				  i++;
				}
				rijndael.rijndaelEncrypt(Key, ref Buff, n);
				n += 16;
			}

			TeleToken.SetLastError(Err.ERROR_SUCCESS);
		  	return true;
		}

		/*****************************************************************************
		Metoda rozszyfruje wczeniej zaszyfrowany blok danych o dowolnym rozmiarze
		We:
			Adres na 16-bajtowy klucz do poczenia
			Obiekt poczenia
			Numer hasa do szyfrowania
			Tablica z danymi
			Ilo danych do rozszyfrowania
			Adres na zmienn do ktrej zostanie wpisana ilo danych po rozszyfrowaniu
		Wy:
			True jesli operacja sie powioda, w przeciwnym wypadku False
			Rozmiar bloku rozszyfrowanego (przez var OSize)
			ustawiony "LastError"
		*****************************************************************************/
		static public bool TVTTAPI_DecryptBlock(byte[] Pass,
												CTVTTAPI_INT_CON Con,
												byte PNr,
												ref byte[] Buff,
												ref uint OSize)
		{
			uint n, i, TSize;
			byte[] Key = new byte[16];
			byte[] Tmp = new byte[16];
			byte[] TBuff;

			if ((Pass == null) || (Con == null)) {
				TeleToken.SetLastError(Err.ERROR_INVALID_DATA);
				return false;
			}

			ZeroMem.CopyTable(ref Key, Buff, 0, 0, 16);
			TVTTAPI_Encrypt(Pass, Con, PNr, ref Key);
			if (TeleToken.GetLastError() != Err.ERROR_SUCCESS) {
				ZeroMem.ZeroMemory(ref Key, 16);
				return false;
			}

			ZeroMem.CopyTable(ref Tmp, Buff, 0, 16, 16);
			rijndael.rijndaelDecrypt(Key, ref Tmp);

			TSize = Tmp[7];
			TSize = TSize << 8;
			TSize = TSize | Tmp[6];
			TSize = TSize << 8;
			TSize = TSize | Tmp[5];
			TSize = TSize << 8;
			TSize = TSize | Tmp[4];

			OSize = Tmp[3];
			OSize = OSize << 8;
			OSize = OSize | Tmp[2];
			OSize = OSize << 8;
			OSize = OSize | Tmp[1];
			OSize = OSize << 8;
			OSize = OSize | Tmp[0];

			TBuff = new byte[TSize-(3*16)];

			ZeroMem.CopyTable(ref TBuff, Buff, 0, 3*16, (int)(TSize-(3*16)));
			ZeroMem.CopyTable(ref Key, Buff, 0, 2*16, 16);
			TVTTAPI_Encrypt(Pass, Con, PNr, ref Key);
			if (TeleToken.GetLastError() != Err.ERROR_SUCCESS) {
				ZeroMem.ZeroMemory(ref Buff, (int) OSize);
				ZeroMem.ZeroMemory(ref TBuff, (int)(TSize-(3*16)));
				ZeroMem.ZeroMemory(ref Key, 16);
				ZeroMem.ZeroMemory(ref Tmp, 16);
				TSize = 0;
				return false;
			}

			n = TSize-(3*16)-16;
			while (n>15) {
				rijndael.rijndaelDecrypt(Key, ref TBuff, n);
				i = 0;
				while (i<16) {
				  TBuff[n+i] ^= TBuff[n+i-16];
				  i++;
				}
				n -= 16;
			}

			ZeroMem.CopyTable(ref Buff, TBuff, 0, 16, (int) OSize);
			ZeroMem.ZeroMemory(ref TBuff, (int)(TSize-(3*16)));

			TeleToken.SetLastError(Err.ERROR_SUCCESS);
		  	return true;
		}
	}
}
